/* Autogenerated */
/* curve description: p224 */
/* requested operations: (all) */
/* s = 0x100000000000000000000000000000000000000000000000000000000 (from "2^224") */
/* c = [(79228162514264337593543950336, 1), (1, -1)] (from "2^96,1;1,-1") */
/* machine_wordsize = 32 (from "32") */
/*                                                                    */
/* NOTE: In addition to the bounds specified above each function, all */
/*   functions synthesized for this Montgomery arithmetic require the */
/*   input to be strictly less than the prime modulus (s-c), and also */
/*   require the input to be in the unique saturated representation.  */
/*   All functions also ensure that these two properties are true of  */
/*   return values.                                                   */

#include <stdint.h>
typedef unsigned char fiat_p224_uint1;
typedef signed char fiat_p224_int1;


/*
 * Input Bounds:
 *   arg1: [0x0 ~> 0x1]
 *   arg2: [0x0 ~> 0xffffffff]
 *   arg3: [0x0 ~> 0xffffffff]
 * Output Bounds:
 *   out1: [0x0 ~> 0xffffffff]
 *   out2: [0x0 ~> 0x1]
 */
static void fiat_p224_addcarryx_u32(uint32_t* out1, fiat_p224_uint1* out2, fiat_p224_uint1 arg1, uint32_t arg2, uint32_t arg3) {
  uint64_t x1 = ((arg1 + (uint64_t)arg2) + arg3);
  uint32_t x2 = (uint32_t)(x1 & UINT32_C(0xffffffff));
  fiat_p224_uint1 x3 = (fiat_p224_uint1)(x1 >> 32);
  *out1 = x2;
  *out2 = x3;
}

/*
 * Input Bounds:
 *   arg1: [0x0 ~> 0x1]
 *   arg2: [0x0 ~> 0xffffffff]
 *   arg3: [0x0 ~> 0xffffffff]
 * Output Bounds:
 *   out1: [0x0 ~> 0xffffffff]
 *   out2: [0x0 ~> 0x1]
 */
static void fiat_p224_subborrowx_u32(uint32_t* out1, fiat_p224_uint1* out2, fiat_p224_uint1 arg1, uint32_t arg2, uint32_t arg3) {
  int64_t x1 = ((arg2 - (int64_t)arg1) - arg3);
  fiat_p224_int1 x2 = (fiat_p224_int1)((uint64_t)x1 >> 32);
  uint32_t x3 = (uint32_t)(x1 & UINT32_C(0xffffffff));
  *out1 = x3;
  *out2 = (fiat_p224_uint1)(0x0 - x2);
}

/*
 * Input Bounds:
 *   arg1: [0x0 ~> 0xffffffff]
 *   arg2: [0x0 ~> 0xffffffff]
 * Output Bounds:
 *   out1: [0x0 ~> 0xffffffff]
 *   out2: [0x0 ~> 0xffffffff]
 */
static void fiat_p224_mulx_u32(uint32_t* out1, uint32_t* out2, uint32_t arg1, uint32_t arg2) {
  uint64_t x1 = ((uint64_t)arg1 * arg2);
  uint32_t x2 = (uint32_t)(x1 & UINT32_C(0xffffffff));
  uint32_t x3 = (uint32_t)(x1 >> 32);
  *out1 = x2;
  *out2 = x3;
}

/*
 * Input Bounds:
 *   arg1: [0x0 ~> 0x1]
 *   arg2: [0x0 ~> 0xffffffff]
 *   arg3: [0x0 ~> 0xffffffff]
 * Output Bounds:
 *   out1: [0x0 ~> 0xffffffff]
 */
static void fiat_p224_cmovznz_u32(uint32_t* out1, fiat_p224_uint1 arg1, uint32_t arg2, uint32_t arg3) {
  fiat_p224_uint1 x1 = (!(!arg1));
  uint32_t x2 = ((fiat_p224_int1)(0x0 - x1) & UINT32_C(0xffffffff));
  uint32_t x3 = ((x2 & arg3) | ((~x2) & arg2));
  *out1 = x3;
}

/*
 * Input Bounds:
 *   arg1: [[0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff]]
 *   arg2: [[0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff]]
 * Output Bounds:
 *   out1: [[0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff]]
 */
static void fiat_p224_mul(uint32_t out1[7], const uint32_t arg1[7], const uint32_t arg2[7]) {
  uint32_t x1 = (arg1[1]);
  uint32_t x2 = (arg1[2]);
  uint32_t x3 = (arg1[3]);
  uint32_t x4 = (arg1[4]);
  uint32_t x5 = (arg1[5]);
  uint32_t x6 = (arg1[6]);
  uint32_t x7 = (arg1[0]);
  uint32_t x8;
  uint32_t x9;
  fiat_p224_mulx_u32(&x8, &x9, x7, (arg2[6]));
  uint32_t x10;
  uint32_t x11;
  fiat_p224_mulx_u32(&x10, &x11, x7, (arg2[5]));
  uint32_t x12;
  uint32_t x13;
  fiat_p224_mulx_u32(&x12, &x13, x7, (arg2[4]));
  uint32_t x14;
  uint32_t x15;
  fiat_p224_mulx_u32(&x14, &x15, x7, (arg2[3]));
  uint32_t x16;
  uint32_t x17;
  fiat_p224_mulx_u32(&x16, &x17, x7, (arg2[2]));
  uint32_t x18;
  uint32_t x19;
  fiat_p224_mulx_u32(&x18, &x19, x7, (arg2[1]));
  uint32_t x20;
  uint32_t x21;
  fiat_p224_mulx_u32(&x20, &x21, x7, (arg2[0]));
  uint32_t x22;
  fiat_p224_uint1 x23;
  fiat_p224_addcarryx_u32(&x22, &x23, 0x0, x18, x21);
  uint32_t x24;
  fiat_p224_uint1 x25;
  fiat_p224_addcarryx_u32(&x24, &x25, x23, x16, x19);
  uint32_t x26;
  fiat_p224_uint1 x27;
  fiat_p224_addcarryx_u32(&x26, &x27, x25, x14, x17);
  uint32_t x28;
  fiat_p224_uint1 x29;
  fiat_p224_addcarryx_u32(&x28, &x29, x27, x12, x15);
  uint32_t x30;
  fiat_p224_uint1 x31;
  fiat_p224_addcarryx_u32(&x30, &x31, x29, x10, x13);
  uint32_t x32;
  fiat_p224_uint1 x33;
  fiat_p224_addcarryx_u32(&x32, &x33, x31, x8, x11);
  uint32_t x34;
  fiat_p224_uint1 x35;
  fiat_p224_addcarryx_u32(&x34, &x35, x33, 0x0, x9);
  uint32_t x36;
  uint32_t x37;
  fiat_p224_mulx_u32(&x36, &x37, x20, UINT32_C(0xffffffff));
  uint32_t x38;
  uint32_t x39;
  fiat_p224_mulx_u32(&x38, &x39, x36, UINT32_C(0xffffffff));
  uint32_t x40;
  uint32_t x41;
  fiat_p224_mulx_u32(&x40, &x41, x36, UINT32_C(0xffffffff));
  uint32_t x42;
  uint32_t x43;
  fiat_p224_mulx_u32(&x42, &x43, x36, UINT32_C(0xffffffff));
  uint32_t x44;
  uint32_t x45;
  fiat_p224_mulx_u32(&x44, &x45, x36, UINT32_C(0xffffffff));
  uint32_t x46;
  fiat_p224_uint1 x47;
  fiat_p224_addcarryx_u32(&x46, &x47, 0x0, x42, x45);
  uint32_t x48;
  fiat_p224_uint1 x49;
  fiat_p224_addcarryx_u32(&x48, &x49, x47, x40, x43);
  uint32_t x50;
  fiat_p224_uint1 x51;
  fiat_p224_addcarryx_u32(&x50, &x51, x49, x38, x41);
  uint32_t x52;
  fiat_p224_uint1 x53;
  fiat_p224_addcarryx_u32(&x52, &x53, x51, 0x0, x39);
  uint32_t x54;
  fiat_p224_uint1 x55;
  fiat_p224_addcarryx_u32(&x54, &x55, 0x0, x36, x20);
  uint32_t x56;
  fiat_p224_uint1 x57;
  fiat_p224_addcarryx_u32(&x56, &x57, x55, 0x0, x22);
  uint32_t x58;
  fiat_p224_uint1 x59;
  fiat_p224_addcarryx_u32(&x58, &x59, x57, 0x0, x24);
  uint32_t x60;
  fiat_p224_uint1 x61;
  fiat_p224_addcarryx_u32(&x60, &x61, x59, x44, x26);
  uint32_t x62;
  fiat_p224_uint1 x63;
  fiat_p224_addcarryx_u32(&x62, &x63, x61, x46, x28);
  uint32_t x64;
  fiat_p224_uint1 x65;
  fiat_p224_addcarryx_u32(&x64, &x65, x63, x48, x30);
  uint32_t x66;
  fiat_p224_uint1 x67;
  fiat_p224_addcarryx_u32(&x66, &x67, x65, x50, x32);
  uint32_t x68;
  fiat_p224_uint1 x69;
  fiat_p224_addcarryx_u32(&x68, &x69, x67, x52, x34);
  uint32_t x70;
  fiat_p224_uint1 x71;
  fiat_p224_addcarryx_u32(&x70, &x71, x69, 0x0, 0x0);
  uint32_t x72;
  uint32_t x73;
  fiat_p224_mulx_u32(&x72, &x73, x1, (arg2[6]));
  uint32_t x74;
  uint32_t x75;
  fiat_p224_mulx_u32(&x74, &x75, x1, (arg2[5]));
  uint32_t x76;
  uint32_t x77;
  fiat_p224_mulx_u32(&x76, &x77, x1, (arg2[4]));
  uint32_t x78;
  uint32_t x79;
  fiat_p224_mulx_u32(&x78, &x79, x1, (arg2[3]));
  uint32_t x80;
  uint32_t x81;
  fiat_p224_mulx_u32(&x80, &x81, x1, (arg2[2]));
  uint32_t x82;
  uint32_t x83;
  fiat_p224_mulx_u32(&x82, &x83, x1, (arg2[1]));
  uint32_t x84;
  uint32_t x85;
  fiat_p224_mulx_u32(&x84, &x85, x1, (arg2[0]));
  uint32_t x86;
  fiat_p224_uint1 x87;
  fiat_p224_addcarryx_u32(&x86, &x87, 0x0, x82, x85);
  uint32_t x88;
  fiat_p224_uint1 x89;
  fiat_p224_addcarryx_u32(&x88, &x89, x87, x80, x83);
  uint32_t x90;
  fiat_p224_uint1 x91;
  fiat_p224_addcarryx_u32(&x90, &x91, x89, x78, x81);
  uint32_t x92;
  fiat_p224_uint1 x93;
  fiat_p224_addcarryx_u32(&x92, &x93, x91, x76, x79);
  uint32_t x94;
  fiat_p224_uint1 x95;
  fiat_p224_addcarryx_u32(&x94, &x95, x93, x74, x77);
  uint32_t x96;
  fiat_p224_uint1 x97;
  fiat_p224_addcarryx_u32(&x96, &x97, x95, x72, x75);
  uint32_t x98;
  fiat_p224_uint1 x99;
  fiat_p224_addcarryx_u32(&x98, &x99, x97, 0x0, x73);
  uint32_t x100;
  fiat_p224_uint1 x101;
  fiat_p224_addcarryx_u32(&x100, &x101, 0x0, x84, x56);
  uint32_t x102;
  fiat_p224_uint1 x103;
  fiat_p224_addcarryx_u32(&x102, &x103, x101, x86, x58);
  uint32_t x104;
  fiat_p224_uint1 x105;
  fiat_p224_addcarryx_u32(&x104, &x105, x103, x88, x60);
  uint32_t x106;
  fiat_p224_uint1 x107;
  fiat_p224_addcarryx_u32(&x106, &x107, x105, x90, x62);
  uint32_t x108;
  fiat_p224_uint1 x109;
  fiat_p224_addcarryx_u32(&x108, &x109, x107, x92, x64);
  uint32_t x110;
  fiat_p224_uint1 x111;
  fiat_p224_addcarryx_u32(&x110, &x111, x109, x94, x66);
  uint32_t x112;
  fiat_p224_uint1 x113;
  fiat_p224_addcarryx_u32(&x112, &x113, x111, x96, x68);
  uint32_t x114;
  fiat_p224_uint1 x115;
  fiat_p224_addcarryx_u32(&x114, &x115, x113, x98, (fiat_p224_uint1)x70);
  uint32_t x116;
  uint32_t x117;
  fiat_p224_mulx_u32(&x116, &x117, x100, UINT32_C(0xffffffff));
  uint32_t x118;
  uint32_t x119;
  fiat_p224_mulx_u32(&x118, &x119, x116, UINT32_C(0xffffffff));
  uint32_t x120;
  uint32_t x121;
  fiat_p224_mulx_u32(&x120, &x121, x116, UINT32_C(0xffffffff));
  uint32_t x122;
  uint32_t x123;
  fiat_p224_mulx_u32(&x122, &x123, x116, UINT32_C(0xffffffff));
  uint32_t x124;
  uint32_t x125;
  fiat_p224_mulx_u32(&x124, &x125, x116, UINT32_C(0xffffffff));
  uint32_t x126;
  fiat_p224_uint1 x127;
  fiat_p224_addcarryx_u32(&x126, &x127, 0x0, x122, x125);
  uint32_t x128;
  fiat_p224_uint1 x129;
  fiat_p224_addcarryx_u32(&x128, &x129, x127, x120, x123);
  uint32_t x130;
  fiat_p224_uint1 x131;
  fiat_p224_addcarryx_u32(&x130, &x131, x129, x118, x121);
  uint32_t x132;
  fiat_p224_uint1 x133;
  fiat_p224_addcarryx_u32(&x132, &x133, x131, 0x0, x119);
  uint32_t x134;
  fiat_p224_uint1 x135;
  fiat_p224_addcarryx_u32(&x134, &x135, 0x0, x116, x100);
  uint32_t x136;
  fiat_p224_uint1 x137;
  fiat_p224_addcarryx_u32(&x136, &x137, x135, 0x0, x102);
  uint32_t x138;
  fiat_p224_uint1 x139;
  fiat_p224_addcarryx_u32(&x138, &x139, x137, 0x0, x104);
  uint32_t x140;
  fiat_p224_uint1 x141;
  fiat_p224_addcarryx_u32(&x140, &x141, x139, x124, x106);
  uint32_t x142;
  fiat_p224_uint1 x143;
  fiat_p224_addcarryx_u32(&x142, &x143, x141, x126, x108);
  uint32_t x144;
  fiat_p224_uint1 x145;
  fiat_p224_addcarryx_u32(&x144, &x145, x143, x128, x110);
  uint32_t x146;
  fiat_p224_uint1 x147;
  fiat_p224_addcarryx_u32(&x146, &x147, x145, x130, x112);
  uint32_t x148;
  fiat_p224_uint1 x149;
  fiat_p224_addcarryx_u32(&x148, &x149, x147, x132, x114);
  uint32_t x150;
  fiat_p224_uint1 x151;
  fiat_p224_addcarryx_u32(&x150, &x151, x149, 0x0, x115);
  uint32_t x152;
  uint32_t x153;
  fiat_p224_mulx_u32(&x152, &x153, x2, (arg2[6]));
  uint32_t x154;
  uint32_t x155;
  fiat_p224_mulx_u32(&x154, &x155, x2, (arg2[5]));
  uint32_t x156;
  uint32_t x157;
  fiat_p224_mulx_u32(&x156, &x157, x2, (arg2[4]));
  uint32_t x158;
  uint32_t x159;
  fiat_p224_mulx_u32(&x158, &x159, x2, (arg2[3]));
  uint32_t x160;
  uint32_t x161;
  fiat_p224_mulx_u32(&x160, &x161, x2, (arg2[2]));
  uint32_t x162;
  uint32_t x163;
  fiat_p224_mulx_u32(&x162, &x163, x2, (arg2[1]));
  uint32_t x164;
  uint32_t x165;
  fiat_p224_mulx_u32(&x164, &x165, x2, (arg2[0]));
  uint32_t x166;
  fiat_p224_uint1 x167;
  fiat_p224_addcarryx_u32(&x166, &x167, 0x0, x162, x165);
  uint32_t x168;
  fiat_p224_uint1 x169;
  fiat_p224_addcarryx_u32(&x168, &x169, x167, x160, x163);
  uint32_t x170;
  fiat_p224_uint1 x171;
  fiat_p224_addcarryx_u32(&x170, &x171, x169, x158, x161);
  uint32_t x172;
  fiat_p224_uint1 x173;
  fiat_p224_addcarryx_u32(&x172, &x173, x171, x156, x159);
  uint32_t x174;
  fiat_p224_uint1 x175;
  fiat_p224_addcarryx_u32(&x174, &x175, x173, x154, x157);
  uint32_t x176;
  fiat_p224_uint1 x177;
  fiat_p224_addcarryx_u32(&x176, &x177, x175, x152, x155);
  uint32_t x178;
  fiat_p224_uint1 x179;
  fiat_p224_addcarryx_u32(&x178, &x179, x177, 0x0, x153);
  uint32_t x180;
  fiat_p224_uint1 x181;
  fiat_p224_addcarryx_u32(&x180, &x181, 0x0, x164, x136);
  uint32_t x182;
  fiat_p224_uint1 x183;
  fiat_p224_addcarryx_u32(&x182, &x183, x181, x166, x138);
  uint32_t x184;
  fiat_p224_uint1 x185;
  fiat_p224_addcarryx_u32(&x184, &x185, x183, x168, x140);
  uint32_t x186;
  fiat_p224_uint1 x187;
  fiat_p224_addcarryx_u32(&x186, &x187, x185, x170, x142);
  uint32_t x188;
  fiat_p224_uint1 x189;
  fiat_p224_addcarryx_u32(&x188, &x189, x187, x172, x144);
  uint32_t x190;
  fiat_p224_uint1 x191;
  fiat_p224_addcarryx_u32(&x190, &x191, x189, x174, x146);
  uint32_t x192;
  fiat_p224_uint1 x193;
  fiat_p224_addcarryx_u32(&x192, &x193, x191, x176, x148);
  uint32_t x194;
  fiat_p224_uint1 x195;
  fiat_p224_addcarryx_u32(&x194, &x195, x193, x178, x150);
  uint32_t x196;
  uint32_t x197;
  fiat_p224_mulx_u32(&x196, &x197, x180, UINT32_C(0xffffffff));
  uint32_t x198;
  uint32_t x199;
  fiat_p224_mulx_u32(&x198, &x199, x196, UINT32_C(0xffffffff));
  uint32_t x200;
  uint32_t x201;
  fiat_p224_mulx_u32(&x200, &x201, x196, UINT32_C(0xffffffff));
  uint32_t x202;
  uint32_t x203;
  fiat_p224_mulx_u32(&x202, &x203, x196, UINT32_C(0xffffffff));
  uint32_t x204;
  uint32_t x205;
  fiat_p224_mulx_u32(&x204, &x205, x196, UINT32_C(0xffffffff));
  uint32_t x206;
  fiat_p224_uint1 x207;
  fiat_p224_addcarryx_u32(&x206, &x207, 0x0, x202, x205);
  uint32_t x208;
  fiat_p224_uint1 x209;
  fiat_p224_addcarryx_u32(&x208, &x209, x207, x200, x203);
  uint32_t x210;
  fiat_p224_uint1 x211;
  fiat_p224_addcarryx_u32(&x210, &x211, x209, x198, x201);
  uint32_t x212;
  fiat_p224_uint1 x213;
  fiat_p224_addcarryx_u32(&x212, &x213, x211, 0x0, x199);
  uint32_t x214;
  fiat_p224_uint1 x215;
  fiat_p224_addcarryx_u32(&x214, &x215, 0x0, x196, x180);
  uint32_t x216;
  fiat_p224_uint1 x217;
  fiat_p224_addcarryx_u32(&x216, &x217, x215, 0x0, x182);
  uint32_t x218;
  fiat_p224_uint1 x219;
  fiat_p224_addcarryx_u32(&x218, &x219, x217, 0x0, x184);
  uint32_t x220;
  fiat_p224_uint1 x221;
  fiat_p224_addcarryx_u32(&x220, &x221, x219, x204, x186);
  uint32_t x222;
  fiat_p224_uint1 x223;
  fiat_p224_addcarryx_u32(&x222, &x223, x221, x206, x188);
  uint32_t x224;
  fiat_p224_uint1 x225;
  fiat_p224_addcarryx_u32(&x224, &x225, x223, x208, x190);
  uint32_t x226;
  fiat_p224_uint1 x227;
  fiat_p224_addcarryx_u32(&x226, &x227, x225, x210, x192);
  uint32_t x228;
  fiat_p224_uint1 x229;
  fiat_p224_addcarryx_u32(&x228, &x229, x227, x212, x194);
  uint32_t x230;
  fiat_p224_uint1 x231;
  fiat_p224_addcarryx_u32(&x230, &x231, x229, 0x0, x195);
  uint32_t x232;
  uint32_t x233;
  fiat_p224_mulx_u32(&x232, &x233, x3, (arg2[6]));
  uint32_t x234;
  uint32_t x235;
  fiat_p224_mulx_u32(&x234, &x235, x3, (arg2[5]));
  uint32_t x236;
  uint32_t x237;
  fiat_p224_mulx_u32(&x236, &x237, x3, (arg2[4]));
  uint32_t x238;
  uint32_t x239;
  fiat_p224_mulx_u32(&x238, &x239, x3, (arg2[3]));
  uint32_t x240;
  uint32_t x241;
  fiat_p224_mulx_u32(&x240, &x241, x3, (arg2[2]));
  uint32_t x242;
  uint32_t x243;
  fiat_p224_mulx_u32(&x242, &x243, x3, (arg2[1]));
  uint32_t x244;
  uint32_t x245;
  fiat_p224_mulx_u32(&x244, &x245, x3, (arg2[0]));
  uint32_t x246;
  fiat_p224_uint1 x247;
  fiat_p224_addcarryx_u32(&x246, &x247, 0x0, x242, x245);
  uint32_t x248;
  fiat_p224_uint1 x249;
  fiat_p224_addcarryx_u32(&x248, &x249, x247, x240, x243);
  uint32_t x250;
  fiat_p224_uint1 x251;
  fiat_p224_addcarryx_u32(&x250, &x251, x249, x238, x241);
  uint32_t x252;
  fiat_p224_uint1 x253;
  fiat_p224_addcarryx_u32(&x252, &x253, x251, x236, x239);
  uint32_t x254;
  fiat_p224_uint1 x255;
  fiat_p224_addcarryx_u32(&x254, &x255, x253, x234, x237);
  uint32_t x256;
  fiat_p224_uint1 x257;
  fiat_p224_addcarryx_u32(&x256, &x257, x255, x232, x235);
  uint32_t x258;
  fiat_p224_uint1 x259;
  fiat_p224_addcarryx_u32(&x258, &x259, x257, 0x0, x233);
  uint32_t x260;
  fiat_p224_uint1 x261;
  fiat_p224_addcarryx_u32(&x260, &x261, 0x0, x244, x216);
  uint32_t x262;
  fiat_p224_uint1 x263;
  fiat_p224_addcarryx_u32(&x262, &x263, x261, x246, x218);
  uint32_t x264;
  fiat_p224_uint1 x265;
  fiat_p224_addcarryx_u32(&x264, &x265, x263, x248, x220);
  uint32_t x266;
  fiat_p224_uint1 x267;
  fiat_p224_addcarryx_u32(&x266, &x267, x265, x250, x222);
  uint32_t x268;
  fiat_p224_uint1 x269;
  fiat_p224_addcarryx_u32(&x268, &x269, x267, x252, x224);
  uint32_t x270;
  fiat_p224_uint1 x271;
  fiat_p224_addcarryx_u32(&x270, &x271, x269, x254, x226);
  uint32_t x272;
  fiat_p224_uint1 x273;
  fiat_p224_addcarryx_u32(&x272, &x273, x271, x256, x228);
  uint32_t x274;
  fiat_p224_uint1 x275;
  fiat_p224_addcarryx_u32(&x274, &x275, x273, x258, x230);
  uint32_t x276;
  uint32_t x277;
  fiat_p224_mulx_u32(&x276, &x277, x260, UINT32_C(0xffffffff));
  uint32_t x278;
  uint32_t x279;
  fiat_p224_mulx_u32(&x278, &x279, x276, UINT32_C(0xffffffff));
  uint32_t x280;
  uint32_t x281;
  fiat_p224_mulx_u32(&x280, &x281, x276, UINT32_C(0xffffffff));
  uint32_t x282;
  uint32_t x283;
  fiat_p224_mulx_u32(&x282, &x283, x276, UINT32_C(0xffffffff));
  uint32_t x284;
  uint32_t x285;
  fiat_p224_mulx_u32(&x284, &x285, x276, UINT32_C(0xffffffff));
  uint32_t x286;
  fiat_p224_uint1 x287;
  fiat_p224_addcarryx_u32(&x286, &x287, 0x0, x282, x285);
  uint32_t x288;
  fiat_p224_uint1 x289;
  fiat_p224_addcarryx_u32(&x288, &x289, x287, x280, x283);
  uint32_t x290;
  fiat_p224_uint1 x291;
  fiat_p224_addcarryx_u32(&x290, &x291, x289, x278, x281);
  uint32_t x292;
  fiat_p224_uint1 x293;
  fiat_p224_addcarryx_u32(&x292, &x293, x291, 0x0, x279);
  uint32_t x294;
  fiat_p224_uint1 x295;
  fiat_p224_addcarryx_u32(&x294, &x295, 0x0, x276, x260);
  uint32_t x296;
  fiat_p224_uint1 x297;
  fiat_p224_addcarryx_u32(&x296, &x297, x295, 0x0, x262);
  uint32_t x298;
  fiat_p224_uint1 x299;
  fiat_p224_addcarryx_u32(&x298, &x299, x297, 0x0, x264);
  uint32_t x300;
  fiat_p224_uint1 x301;
  fiat_p224_addcarryx_u32(&x300, &x301, x299, x284, x266);
  uint32_t x302;
  fiat_p224_uint1 x303;
  fiat_p224_addcarryx_u32(&x302, &x303, x301, x286, x268);
  uint32_t x304;
  fiat_p224_uint1 x305;
  fiat_p224_addcarryx_u32(&x304, &x305, x303, x288, x270);
  uint32_t x306;
  fiat_p224_uint1 x307;
  fiat_p224_addcarryx_u32(&x306, &x307, x305, x290, x272);
  uint32_t x308;
  fiat_p224_uint1 x309;
  fiat_p224_addcarryx_u32(&x308, &x309, x307, x292, x274);
  uint32_t x310;
  fiat_p224_uint1 x311;
  fiat_p224_addcarryx_u32(&x310, &x311, x309, 0x0, x275);
  uint32_t x312;
  uint32_t x313;
  fiat_p224_mulx_u32(&x312, &x313, x4, (arg2[6]));
  uint32_t x314;
  uint32_t x315;
  fiat_p224_mulx_u32(&x314, &x315, x4, (arg2[5]));
  uint32_t x316;
  uint32_t x317;
  fiat_p224_mulx_u32(&x316, &x317, x4, (arg2[4]));
  uint32_t x318;
  uint32_t x319;
  fiat_p224_mulx_u32(&x318, &x319, x4, (arg2[3]));
  uint32_t x320;
  uint32_t x321;
  fiat_p224_mulx_u32(&x320, &x321, x4, (arg2[2]));
  uint32_t x322;
  uint32_t x323;
  fiat_p224_mulx_u32(&x322, &x323, x4, (arg2[1]));
  uint32_t x324;
  uint32_t x325;
  fiat_p224_mulx_u32(&x324, &x325, x4, (arg2[0]));
  uint32_t x326;
  fiat_p224_uint1 x327;
  fiat_p224_addcarryx_u32(&x326, &x327, 0x0, x322, x325);
  uint32_t x328;
  fiat_p224_uint1 x329;
  fiat_p224_addcarryx_u32(&x328, &x329, x327, x320, x323);
  uint32_t x330;
  fiat_p224_uint1 x331;
  fiat_p224_addcarryx_u32(&x330, &x331, x329, x318, x321);
  uint32_t x332;
  fiat_p224_uint1 x333;
  fiat_p224_addcarryx_u32(&x332, &x333, x331, x316, x319);
  uint32_t x334;
  fiat_p224_uint1 x335;
  fiat_p224_addcarryx_u32(&x334, &x335, x333, x314, x317);
  uint32_t x336;
  fiat_p224_uint1 x337;
  fiat_p224_addcarryx_u32(&x336, &x337, x335, x312, x315);
  uint32_t x338;
  fiat_p224_uint1 x339;
  fiat_p224_addcarryx_u32(&x338, &x339, x337, 0x0, x313);
  uint32_t x340;
  fiat_p224_uint1 x341;
  fiat_p224_addcarryx_u32(&x340, &x341, 0x0, x324, x296);
  uint32_t x342;
  fiat_p224_uint1 x343;
  fiat_p224_addcarryx_u32(&x342, &x343, x341, x326, x298);
  uint32_t x344;
  fiat_p224_uint1 x345;
  fiat_p224_addcarryx_u32(&x344, &x345, x343, x328, x300);
  uint32_t x346;
  fiat_p224_uint1 x347;
  fiat_p224_addcarryx_u32(&x346, &x347, x345, x330, x302);
  uint32_t x348;
  fiat_p224_uint1 x349;
  fiat_p224_addcarryx_u32(&x348, &x349, x347, x332, x304);
  uint32_t x350;
  fiat_p224_uint1 x351;
  fiat_p224_addcarryx_u32(&x350, &x351, x349, x334, x306);
  uint32_t x352;
  fiat_p224_uint1 x353;
  fiat_p224_addcarryx_u32(&x352, &x353, x351, x336, x308);
  uint32_t x354;
  fiat_p224_uint1 x355;
  fiat_p224_addcarryx_u32(&x354, &x355, x353, x338, x310);
  uint32_t x356;
  uint32_t x357;
  fiat_p224_mulx_u32(&x356, &x357, x340, UINT32_C(0xffffffff));
  uint32_t x358;
  uint32_t x359;
  fiat_p224_mulx_u32(&x358, &x359, x356, UINT32_C(0xffffffff));
  uint32_t x360;
  uint32_t x361;
  fiat_p224_mulx_u32(&x360, &x361, x356, UINT32_C(0xffffffff));
  uint32_t x362;
  uint32_t x363;
  fiat_p224_mulx_u32(&x362, &x363, x356, UINT32_C(0xffffffff));
  uint32_t x364;
  uint32_t x365;
  fiat_p224_mulx_u32(&x364, &x365, x356, UINT32_C(0xffffffff));
  uint32_t x366;
  fiat_p224_uint1 x367;
  fiat_p224_addcarryx_u32(&x366, &x367, 0x0, x362, x365);
  uint32_t x368;
  fiat_p224_uint1 x369;
  fiat_p224_addcarryx_u32(&x368, &x369, x367, x360, x363);
  uint32_t x370;
  fiat_p224_uint1 x371;
  fiat_p224_addcarryx_u32(&x370, &x371, x369, x358, x361);
  uint32_t x372;
  fiat_p224_uint1 x373;
  fiat_p224_addcarryx_u32(&x372, &x373, x371, 0x0, x359);
  uint32_t x374;
  fiat_p224_uint1 x375;
  fiat_p224_addcarryx_u32(&x374, &x375, 0x0, x356, x340);
  uint32_t x376;
  fiat_p224_uint1 x377;
  fiat_p224_addcarryx_u32(&x376, &x377, x375, 0x0, x342);
  uint32_t x378;
  fiat_p224_uint1 x379;
  fiat_p224_addcarryx_u32(&x378, &x379, x377, 0x0, x344);
  uint32_t x380;
  fiat_p224_uint1 x381;
  fiat_p224_addcarryx_u32(&x380, &x381, x379, x364, x346);
  uint32_t x382;
  fiat_p224_uint1 x383;
  fiat_p224_addcarryx_u32(&x382, &x383, x381, x366, x348);
  uint32_t x384;
  fiat_p224_uint1 x385;
  fiat_p224_addcarryx_u32(&x384, &x385, x383, x368, x350);
  uint32_t x386;
  fiat_p224_uint1 x387;
  fiat_p224_addcarryx_u32(&x386, &x387, x385, x370, x352);
  uint32_t x388;
  fiat_p224_uint1 x389;
  fiat_p224_addcarryx_u32(&x388, &x389, x387, x372, x354);
  uint32_t x390;
  fiat_p224_uint1 x391;
  fiat_p224_addcarryx_u32(&x390, &x391, x389, 0x0, x355);
  uint32_t x392;
  uint32_t x393;
  fiat_p224_mulx_u32(&x392, &x393, x5, (arg2[6]));
  uint32_t x394;
  uint32_t x395;
  fiat_p224_mulx_u32(&x394, &x395, x5, (arg2[5]));
  uint32_t x396;
  uint32_t x397;
  fiat_p224_mulx_u32(&x396, &x397, x5, (arg2[4]));
  uint32_t x398;
  uint32_t x399;
  fiat_p224_mulx_u32(&x398, &x399, x5, (arg2[3]));
  uint32_t x400;
  uint32_t x401;
  fiat_p224_mulx_u32(&x400, &x401, x5, (arg2[2]));
  uint32_t x402;
  uint32_t x403;
  fiat_p224_mulx_u32(&x402, &x403, x5, (arg2[1]));
  uint32_t x404;
  uint32_t x405;
  fiat_p224_mulx_u32(&x404, &x405, x5, (arg2[0]));
  uint32_t x406;
  fiat_p224_uint1 x407;
  fiat_p224_addcarryx_u32(&x406, &x407, 0x0, x402, x405);
  uint32_t x408;
  fiat_p224_uint1 x409;
  fiat_p224_addcarryx_u32(&x408, &x409, x407, x400, x403);
  uint32_t x410;
  fiat_p224_uint1 x411;
  fiat_p224_addcarryx_u32(&x410, &x411, x409, x398, x401);
  uint32_t x412;
  fiat_p224_uint1 x413;
  fiat_p224_addcarryx_u32(&x412, &x413, x411, x396, x399);
  uint32_t x414;
  fiat_p224_uint1 x415;
  fiat_p224_addcarryx_u32(&x414, &x415, x413, x394, x397);
  uint32_t x416;
  fiat_p224_uint1 x417;
  fiat_p224_addcarryx_u32(&x416, &x417, x415, x392, x395);
  uint32_t x418;
  fiat_p224_uint1 x419;
  fiat_p224_addcarryx_u32(&x418, &x419, x417, 0x0, x393);
  uint32_t x420;
  fiat_p224_uint1 x421;
  fiat_p224_addcarryx_u32(&x420, &x421, 0x0, x404, x376);
  uint32_t x422;
  fiat_p224_uint1 x423;
  fiat_p224_addcarryx_u32(&x422, &x423, x421, x406, x378);
  uint32_t x424;
  fiat_p224_uint1 x425;
  fiat_p224_addcarryx_u32(&x424, &x425, x423, x408, x380);
  uint32_t x426;
  fiat_p224_uint1 x427;
  fiat_p224_addcarryx_u32(&x426, &x427, x425, x410, x382);
  uint32_t x428;
  fiat_p224_uint1 x429;
  fiat_p224_addcarryx_u32(&x428, &x429, x427, x412, x384);
  uint32_t x430;
  fiat_p224_uint1 x431;
  fiat_p224_addcarryx_u32(&x430, &x431, x429, x414, x386);
  uint32_t x432;
  fiat_p224_uint1 x433;
  fiat_p224_addcarryx_u32(&x432, &x433, x431, x416, x388);
  uint32_t x434;
  fiat_p224_uint1 x435;
  fiat_p224_addcarryx_u32(&x434, &x435, x433, x418, x390);
  uint32_t x436;
  uint32_t x437;
  fiat_p224_mulx_u32(&x436, &x437, x420, UINT32_C(0xffffffff));
  uint32_t x438;
  uint32_t x439;
  fiat_p224_mulx_u32(&x438, &x439, x436, UINT32_C(0xffffffff));
  uint32_t x440;
  uint32_t x441;
  fiat_p224_mulx_u32(&x440, &x441, x436, UINT32_C(0xffffffff));
  uint32_t x442;
  uint32_t x443;
  fiat_p224_mulx_u32(&x442, &x443, x436, UINT32_C(0xffffffff));
  uint32_t x444;
  uint32_t x445;
  fiat_p224_mulx_u32(&x444, &x445, x436, UINT32_C(0xffffffff));
  uint32_t x446;
  fiat_p224_uint1 x447;
  fiat_p224_addcarryx_u32(&x446, &x447, 0x0, x442, x445);
  uint32_t x448;
  fiat_p224_uint1 x449;
  fiat_p224_addcarryx_u32(&x448, &x449, x447, x440, x443);
  uint32_t x450;
  fiat_p224_uint1 x451;
  fiat_p224_addcarryx_u32(&x450, &x451, x449, x438, x441);
  uint32_t x452;
  fiat_p224_uint1 x453;
  fiat_p224_addcarryx_u32(&x452, &x453, x451, 0x0, x439);
  uint32_t x454;
  fiat_p224_uint1 x455;
  fiat_p224_addcarryx_u32(&x454, &x455, 0x0, x436, x420);
  uint32_t x456;
  fiat_p224_uint1 x457;
  fiat_p224_addcarryx_u32(&x456, &x457, x455, 0x0, x422);
  uint32_t x458;
  fiat_p224_uint1 x459;
  fiat_p224_addcarryx_u32(&x458, &x459, x457, 0x0, x424);
  uint32_t x460;
  fiat_p224_uint1 x461;
  fiat_p224_addcarryx_u32(&x460, &x461, x459, x444, x426);
  uint32_t x462;
  fiat_p224_uint1 x463;
  fiat_p224_addcarryx_u32(&x462, &x463, x461, x446, x428);
  uint32_t x464;
  fiat_p224_uint1 x465;
  fiat_p224_addcarryx_u32(&x464, &x465, x463, x448, x430);
  uint32_t x466;
  fiat_p224_uint1 x467;
  fiat_p224_addcarryx_u32(&x466, &x467, x465, x450, x432);
  uint32_t x468;
  fiat_p224_uint1 x469;
  fiat_p224_addcarryx_u32(&x468, &x469, x467, x452, x434);
  uint32_t x470;
  fiat_p224_uint1 x471;
  fiat_p224_addcarryx_u32(&x470, &x471, x469, 0x0, x435);
  uint32_t x472;
  uint32_t x473;
  fiat_p224_mulx_u32(&x472, &x473, x6, (arg2[6]));
  uint32_t x474;
  uint32_t x475;
  fiat_p224_mulx_u32(&x474, &x475, x6, (arg2[5]));
  uint32_t x476;
  uint32_t x477;
  fiat_p224_mulx_u32(&x476, &x477, x6, (arg2[4]));
  uint32_t x478;
  uint32_t x479;
  fiat_p224_mulx_u32(&x478, &x479, x6, (arg2[3]));
  uint32_t x480;
  uint32_t x481;
  fiat_p224_mulx_u32(&x480, &x481, x6, (arg2[2]));
  uint32_t x482;
  uint32_t x483;
  fiat_p224_mulx_u32(&x482, &x483, x6, (arg2[1]));
  uint32_t x484;
  uint32_t x485;
  fiat_p224_mulx_u32(&x484, &x485, x6, (arg2[0]));
  uint32_t x486;
  fiat_p224_uint1 x487;
  fiat_p224_addcarryx_u32(&x486, &x487, 0x0, x482, x485);
  uint32_t x488;
  fiat_p224_uint1 x489;
  fiat_p224_addcarryx_u32(&x488, &x489, x487, x480, x483);
  uint32_t x490;
  fiat_p224_uint1 x491;
  fiat_p224_addcarryx_u32(&x490, &x491, x489, x478, x481);
  uint32_t x492;
  fiat_p224_uint1 x493;
  fiat_p224_addcarryx_u32(&x492, &x493, x491, x476, x479);
  uint32_t x494;
  fiat_p224_uint1 x495;
  fiat_p224_addcarryx_u32(&x494, &x495, x493, x474, x477);
  uint32_t x496;
  fiat_p224_uint1 x497;
  fiat_p224_addcarryx_u32(&x496, &x497, x495, x472, x475);
  uint32_t x498;
  fiat_p224_uint1 x499;
  fiat_p224_addcarryx_u32(&x498, &x499, x497, 0x0, x473);
  uint32_t x500;
  fiat_p224_uint1 x501;
  fiat_p224_addcarryx_u32(&x500, &x501, 0x0, x484, x456);
  uint32_t x502;
  fiat_p224_uint1 x503;
  fiat_p224_addcarryx_u32(&x502, &x503, x501, x486, x458);
  uint32_t x504;
  fiat_p224_uint1 x505;
  fiat_p224_addcarryx_u32(&x504, &x505, x503, x488, x460);
  uint32_t x506;
  fiat_p224_uint1 x507;
  fiat_p224_addcarryx_u32(&x506, &x507, x505, x490, x462);
  uint32_t x508;
  fiat_p224_uint1 x509;
  fiat_p224_addcarryx_u32(&x508, &x509, x507, x492, x464);
  uint32_t x510;
  fiat_p224_uint1 x511;
  fiat_p224_addcarryx_u32(&x510, &x511, x509, x494, x466);
  uint32_t x512;
  fiat_p224_uint1 x513;
  fiat_p224_addcarryx_u32(&x512, &x513, x511, x496, x468);
  uint32_t x514;
  fiat_p224_uint1 x515;
  fiat_p224_addcarryx_u32(&x514, &x515, x513, x498, x470);
  uint32_t x516;
  uint32_t x517;
  fiat_p224_mulx_u32(&x516, &x517, x500, UINT32_C(0xffffffff));
  uint32_t x518;
  uint32_t x519;
  fiat_p224_mulx_u32(&x518, &x519, x516, UINT32_C(0xffffffff));
  uint32_t x520;
  uint32_t x521;
  fiat_p224_mulx_u32(&x520, &x521, x516, UINT32_C(0xffffffff));
  uint32_t x522;
  uint32_t x523;
  fiat_p224_mulx_u32(&x522, &x523, x516, UINT32_C(0xffffffff));
  uint32_t x524;
  uint32_t x525;
  fiat_p224_mulx_u32(&x524, &x525, x516, UINT32_C(0xffffffff));
  uint32_t x526;
  fiat_p224_uint1 x527;
  fiat_p224_addcarryx_u32(&x526, &x527, 0x0, x522, x525);
  uint32_t x528;
  fiat_p224_uint1 x529;
  fiat_p224_addcarryx_u32(&x528, &x529, x527, x520, x523);
  uint32_t x530;
  fiat_p224_uint1 x531;
  fiat_p224_addcarryx_u32(&x530, &x531, x529, x518, x521);
  uint32_t x532;
  fiat_p224_uint1 x533;
  fiat_p224_addcarryx_u32(&x532, &x533, x531, 0x0, x519);
  uint32_t x534;
  fiat_p224_uint1 x535;
  fiat_p224_addcarryx_u32(&x534, &x535, 0x0, x516, x500);
  uint32_t x536;
  fiat_p224_uint1 x537;
  fiat_p224_addcarryx_u32(&x536, &x537, x535, 0x0, x502);
  uint32_t x538;
  fiat_p224_uint1 x539;
  fiat_p224_addcarryx_u32(&x538, &x539, x537, 0x0, x504);
  uint32_t x540;
  fiat_p224_uint1 x541;
  fiat_p224_addcarryx_u32(&x540, &x541, x539, x524, x506);
  uint32_t x542;
  fiat_p224_uint1 x543;
  fiat_p224_addcarryx_u32(&x542, &x543, x541, x526, x508);
  uint32_t x544;
  fiat_p224_uint1 x545;
  fiat_p224_addcarryx_u32(&x544, &x545, x543, x528, x510);
  uint32_t x546;
  fiat_p224_uint1 x547;
  fiat_p224_addcarryx_u32(&x546, &x547, x545, x530, x512);
  uint32_t x548;
  fiat_p224_uint1 x549;
  fiat_p224_addcarryx_u32(&x548, &x549, x547, x532, x514);
  uint32_t x550;
  fiat_p224_uint1 x551;
  fiat_p224_addcarryx_u32(&x550, &x551, x549, 0x0, x515);
  uint32_t x552;
  fiat_p224_uint1 x553;
  fiat_p224_subborrowx_u32(&x552, &x553, 0x0, x536, 0x1);
  uint32_t x554;
  fiat_p224_uint1 x555;
  fiat_p224_subborrowx_u32(&x554, &x555, x553, x538, 0x0);
  uint32_t x556;
  fiat_p224_uint1 x557;
  fiat_p224_subborrowx_u32(&x556, &x557, x555, x540, 0x0);
  uint32_t x558;
  fiat_p224_uint1 x559;
  fiat_p224_subborrowx_u32(&x558, &x559, x557, x542, UINT32_C(0xffffffff));
  uint32_t x560;
  fiat_p224_uint1 x561;
  fiat_p224_subborrowx_u32(&x560, &x561, x559, x544, UINT32_C(0xffffffff));
  uint32_t x562;
  fiat_p224_uint1 x563;
  fiat_p224_subborrowx_u32(&x562, &x563, x561, x546, UINT32_C(0xffffffff));
  uint32_t x564;
  fiat_p224_uint1 x565;
  fiat_p224_subborrowx_u32(&x564, &x565, x563, x548, UINT32_C(0xffffffff));
  uint32_t x566;
  fiat_p224_uint1 x567;
  fiat_p224_subborrowx_u32(&x566, &x567, x565, x550, 0x0);
  uint32_t x568;
  fiat_p224_cmovznz_u32(&x568, x567, x552, x536);
  uint32_t x569;
  fiat_p224_cmovznz_u32(&x569, x567, x554, x538);
  uint32_t x570;
  fiat_p224_cmovznz_u32(&x570, x567, x556, x540);
  uint32_t x571;
  fiat_p224_cmovznz_u32(&x571, x567, x558, x542);
  uint32_t x572;
  fiat_p224_cmovznz_u32(&x572, x567, x560, x544);
  uint32_t x573;
  fiat_p224_cmovznz_u32(&x573, x567, x562, x546);
  uint32_t x574;
  fiat_p224_cmovznz_u32(&x574, x567, x564, x548);
  out1[0] = x568;
  out1[1] = x569;
  out1[2] = x570;
  out1[3] = x571;
  out1[4] = x572;
  out1[5] = x573;
  out1[6] = x574;
}

/*
 * Input Bounds:
 *   arg1: [[0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff]]
 * Output Bounds:
 *   out1: [[0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff]]
 */
static void fiat_p224_square(uint32_t out1[7], const uint32_t arg1[7]) {
  uint32_t x1 = (arg1[1]);
  uint32_t x2 = (arg1[2]);
  uint32_t x3 = (arg1[3]);
  uint32_t x4 = (arg1[4]);
  uint32_t x5 = (arg1[5]);
  uint32_t x6 = (arg1[6]);
  uint32_t x7 = (arg1[0]);
  uint32_t x8;
  uint32_t x9;
  fiat_p224_mulx_u32(&x8, &x9, x7, (arg1[6]));
  uint32_t x10;
  uint32_t x11;
  fiat_p224_mulx_u32(&x10, &x11, x7, (arg1[5]));
  uint32_t x12;
  uint32_t x13;
  fiat_p224_mulx_u32(&x12, &x13, x7, (arg1[4]));
  uint32_t x14;
  uint32_t x15;
  fiat_p224_mulx_u32(&x14, &x15, x7, (arg1[3]));
  uint32_t x16;
  uint32_t x17;
  fiat_p224_mulx_u32(&x16, &x17, x7, (arg1[2]));
  uint32_t x18;
  uint32_t x19;
  fiat_p224_mulx_u32(&x18, &x19, x7, (arg1[1]));
  uint32_t x20;
  uint32_t x21;
  fiat_p224_mulx_u32(&x20, &x21, x7, (arg1[0]));
  uint32_t x22;
  fiat_p224_uint1 x23;
  fiat_p224_addcarryx_u32(&x22, &x23, 0x0, x18, x21);
  uint32_t x24;
  fiat_p224_uint1 x25;
  fiat_p224_addcarryx_u32(&x24, &x25, x23, x16, x19);
  uint32_t x26;
  fiat_p224_uint1 x27;
  fiat_p224_addcarryx_u32(&x26, &x27, x25, x14, x17);
  uint32_t x28;
  fiat_p224_uint1 x29;
  fiat_p224_addcarryx_u32(&x28, &x29, x27, x12, x15);
  uint32_t x30;
  fiat_p224_uint1 x31;
  fiat_p224_addcarryx_u32(&x30, &x31, x29, x10, x13);
  uint32_t x32;
  fiat_p224_uint1 x33;
  fiat_p224_addcarryx_u32(&x32, &x33, x31, x8, x11);
  uint32_t x34;
  fiat_p224_uint1 x35;
  fiat_p224_addcarryx_u32(&x34, &x35, x33, 0x0, x9);
  uint32_t x36;
  uint32_t x37;
  fiat_p224_mulx_u32(&x36, &x37, x20, UINT32_C(0xffffffff));
  uint32_t x38;
  uint32_t x39;
  fiat_p224_mulx_u32(&x38, &x39, x36, UINT32_C(0xffffffff));
  uint32_t x40;
  uint32_t x41;
  fiat_p224_mulx_u32(&x40, &x41, x36, UINT32_C(0xffffffff));
  uint32_t x42;
  uint32_t x43;
  fiat_p224_mulx_u32(&x42, &x43, x36, UINT32_C(0xffffffff));
  uint32_t x44;
  uint32_t x45;
  fiat_p224_mulx_u32(&x44, &x45, x36, UINT32_C(0xffffffff));
  uint32_t x46;
  fiat_p224_uint1 x47;
  fiat_p224_addcarryx_u32(&x46, &x47, 0x0, x42, x45);
  uint32_t x48;
  fiat_p224_uint1 x49;
  fiat_p224_addcarryx_u32(&x48, &x49, x47, x40, x43);
  uint32_t x50;
  fiat_p224_uint1 x51;
  fiat_p224_addcarryx_u32(&x50, &x51, x49, x38, x41);
  uint32_t x52;
  fiat_p224_uint1 x53;
  fiat_p224_addcarryx_u32(&x52, &x53, x51, 0x0, x39);
  uint32_t x54;
  fiat_p224_uint1 x55;
  fiat_p224_addcarryx_u32(&x54, &x55, 0x0, x36, x20);
  uint32_t x56;
  fiat_p224_uint1 x57;
  fiat_p224_addcarryx_u32(&x56, &x57, x55, 0x0, x22);
  uint32_t x58;
  fiat_p224_uint1 x59;
  fiat_p224_addcarryx_u32(&x58, &x59, x57, 0x0, x24);
  uint32_t x60;
  fiat_p224_uint1 x61;
  fiat_p224_addcarryx_u32(&x60, &x61, x59, x44, x26);
  uint32_t x62;
  fiat_p224_uint1 x63;
  fiat_p224_addcarryx_u32(&x62, &x63, x61, x46, x28);
  uint32_t x64;
  fiat_p224_uint1 x65;
  fiat_p224_addcarryx_u32(&x64, &x65, x63, x48, x30);
  uint32_t x66;
  fiat_p224_uint1 x67;
  fiat_p224_addcarryx_u32(&x66, &x67, x65, x50, x32);
  uint32_t x68;
  fiat_p224_uint1 x69;
  fiat_p224_addcarryx_u32(&x68, &x69, x67, x52, x34);
  uint32_t x70;
  fiat_p224_uint1 x71;
  fiat_p224_addcarryx_u32(&x70, &x71, x69, 0x0, 0x0);
  uint32_t x72;
  uint32_t x73;
  fiat_p224_mulx_u32(&x72, &x73, x1, (arg1[6]));
  uint32_t x74;
  uint32_t x75;
  fiat_p224_mulx_u32(&x74, &x75, x1, (arg1[5]));
  uint32_t x76;
  uint32_t x77;
  fiat_p224_mulx_u32(&x76, &x77, x1, (arg1[4]));
  uint32_t x78;
  uint32_t x79;
  fiat_p224_mulx_u32(&x78, &x79, x1, (arg1[3]));
  uint32_t x80;
  uint32_t x81;
  fiat_p224_mulx_u32(&x80, &x81, x1, (arg1[2]));
  uint32_t x82;
  uint32_t x83;
  fiat_p224_mulx_u32(&x82, &x83, x1, (arg1[1]));
  uint32_t x84;
  uint32_t x85;
  fiat_p224_mulx_u32(&x84, &x85, x1, (arg1[0]));
  uint32_t x86;
  fiat_p224_uint1 x87;
  fiat_p224_addcarryx_u32(&x86, &x87, 0x0, x82, x85);
  uint32_t x88;
  fiat_p224_uint1 x89;
  fiat_p224_addcarryx_u32(&x88, &x89, x87, x80, x83);
  uint32_t x90;
  fiat_p224_uint1 x91;
  fiat_p224_addcarryx_u32(&x90, &x91, x89, x78, x81);
  uint32_t x92;
  fiat_p224_uint1 x93;
  fiat_p224_addcarryx_u32(&x92, &x93, x91, x76, x79);
  uint32_t x94;
  fiat_p224_uint1 x95;
  fiat_p224_addcarryx_u32(&x94, &x95, x93, x74, x77);
  uint32_t x96;
  fiat_p224_uint1 x97;
  fiat_p224_addcarryx_u32(&x96, &x97, x95, x72, x75);
  uint32_t x98;
  fiat_p224_uint1 x99;
  fiat_p224_addcarryx_u32(&x98, &x99, x97, 0x0, x73);
  uint32_t x100;
  fiat_p224_uint1 x101;
  fiat_p224_addcarryx_u32(&x100, &x101, 0x0, x84, x56);
  uint32_t x102;
  fiat_p224_uint1 x103;
  fiat_p224_addcarryx_u32(&x102, &x103, x101, x86, x58);
  uint32_t x104;
  fiat_p224_uint1 x105;
  fiat_p224_addcarryx_u32(&x104, &x105, x103, x88, x60);
  uint32_t x106;
  fiat_p224_uint1 x107;
  fiat_p224_addcarryx_u32(&x106, &x107, x105, x90, x62);
  uint32_t x108;
  fiat_p224_uint1 x109;
  fiat_p224_addcarryx_u32(&x108, &x109, x107, x92, x64);
  uint32_t x110;
  fiat_p224_uint1 x111;
  fiat_p224_addcarryx_u32(&x110, &x111, x109, x94, x66);
  uint32_t x112;
  fiat_p224_uint1 x113;
  fiat_p224_addcarryx_u32(&x112, &x113, x111, x96, x68);
  uint32_t x114;
  fiat_p224_uint1 x115;
  fiat_p224_addcarryx_u32(&x114, &x115, x113, x98, (fiat_p224_uint1)x70);
  uint32_t x116;
  uint32_t x117;
  fiat_p224_mulx_u32(&x116, &x117, x100, UINT32_C(0xffffffff));
  uint32_t x118;
  uint32_t x119;
  fiat_p224_mulx_u32(&x118, &x119, x116, UINT32_C(0xffffffff));
  uint32_t x120;
  uint32_t x121;
  fiat_p224_mulx_u32(&x120, &x121, x116, UINT32_C(0xffffffff));
  uint32_t x122;
  uint32_t x123;
  fiat_p224_mulx_u32(&x122, &x123, x116, UINT32_C(0xffffffff));
  uint32_t x124;
  uint32_t x125;
  fiat_p224_mulx_u32(&x124, &x125, x116, UINT32_C(0xffffffff));
  uint32_t x126;
  fiat_p224_uint1 x127;
  fiat_p224_addcarryx_u32(&x126, &x127, 0x0, x122, x125);
  uint32_t x128;
  fiat_p224_uint1 x129;
  fiat_p224_addcarryx_u32(&x128, &x129, x127, x120, x123);
  uint32_t x130;
  fiat_p224_uint1 x131;
  fiat_p224_addcarryx_u32(&x130, &x131, x129, x118, x121);
  uint32_t x132;
  fiat_p224_uint1 x133;
  fiat_p224_addcarryx_u32(&x132, &x133, x131, 0x0, x119);
  uint32_t x134;
  fiat_p224_uint1 x135;
  fiat_p224_addcarryx_u32(&x134, &x135, 0x0, x116, x100);
  uint32_t x136;
  fiat_p224_uint1 x137;
  fiat_p224_addcarryx_u32(&x136, &x137, x135, 0x0, x102);
  uint32_t x138;
  fiat_p224_uint1 x139;
  fiat_p224_addcarryx_u32(&x138, &x139, x137, 0x0, x104);
  uint32_t x140;
  fiat_p224_uint1 x141;
  fiat_p224_addcarryx_u32(&x140, &x141, x139, x124, x106);
  uint32_t x142;
  fiat_p224_uint1 x143;
  fiat_p224_addcarryx_u32(&x142, &x143, x141, x126, x108);
  uint32_t x144;
  fiat_p224_uint1 x145;
  fiat_p224_addcarryx_u32(&x144, &x145, x143, x128, x110);
  uint32_t x146;
  fiat_p224_uint1 x147;
  fiat_p224_addcarryx_u32(&x146, &x147, x145, x130, x112);
  uint32_t x148;
  fiat_p224_uint1 x149;
  fiat_p224_addcarryx_u32(&x148, &x149, x147, x132, x114);
  uint32_t x150;
  fiat_p224_uint1 x151;
  fiat_p224_addcarryx_u32(&x150, &x151, x149, 0x0, x115);
  uint32_t x152;
  uint32_t x153;
  fiat_p224_mulx_u32(&x152, &x153, x2, (arg1[6]));
  uint32_t x154;
  uint32_t x155;
  fiat_p224_mulx_u32(&x154, &x155, x2, (arg1[5]));
  uint32_t x156;
  uint32_t x157;
  fiat_p224_mulx_u32(&x156, &x157, x2, (arg1[4]));
  uint32_t x158;
  uint32_t x159;
  fiat_p224_mulx_u32(&x158, &x159, x2, (arg1[3]));
  uint32_t x160;
  uint32_t x161;
  fiat_p224_mulx_u32(&x160, &x161, x2, (arg1[2]));
  uint32_t x162;
  uint32_t x163;
  fiat_p224_mulx_u32(&x162, &x163, x2, (arg1[1]));
  uint32_t x164;
  uint32_t x165;
  fiat_p224_mulx_u32(&x164, &x165, x2, (arg1[0]));
  uint32_t x166;
  fiat_p224_uint1 x167;
  fiat_p224_addcarryx_u32(&x166, &x167, 0x0, x162, x165);
  uint32_t x168;
  fiat_p224_uint1 x169;
  fiat_p224_addcarryx_u32(&x168, &x169, x167, x160, x163);
  uint32_t x170;
  fiat_p224_uint1 x171;
  fiat_p224_addcarryx_u32(&x170, &x171, x169, x158, x161);
  uint32_t x172;
  fiat_p224_uint1 x173;
  fiat_p224_addcarryx_u32(&x172, &x173, x171, x156, x159);
  uint32_t x174;
  fiat_p224_uint1 x175;
  fiat_p224_addcarryx_u32(&x174, &x175, x173, x154, x157);
  uint32_t x176;
  fiat_p224_uint1 x177;
  fiat_p224_addcarryx_u32(&x176, &x177, x175, x152, x155);
  uint32_t x178;
  fiat_p224_uint1 x179;
  fiat_p224_addcarryx_u32(&x178, &x179, x177, 0x0, x153);
  uint32_t x180;
  fiat_p224_uint1 x181;
  fiat_p224_addcarryx_u32(&x180, &x181, 0x0, x164, x136);
  uint32_t x182;
  fiat_p224_uint1 x183;
  fiat_p224_addcarryx_u32(&x182, &x183, x181, x166, x138);
  uint32_t x184;
  fiat_p224_uint1 x185;
  fiat_p224_addcarryx_u32(&x184, &x185, x183, x168, x140);
  uint32_t x186;
  fiat_p224_uint1 x187;
  fiat_p224_addcarryx_u32(&x186, &x187, x185, x170, x142);
  uint32_t x188;
  fiat_p224_uint1 x189;
  fiat_p224_addcarryx_u32(&x188, &x189, x187, x172, x144);
  uint32_t x190;
  fiat_p224_uint1 x191;
  fiat_p224_addcarryx_u32(&x190, &x191, x189, x174, x146);
  uint32_t x192;
  fiat_p224_uint1 x193;
  fiat_p224_addcarryx_u32(&x192, &x193, x191, x176, x148);
  uint32_t x194;
  fiat_p224_uint1 x195;
  fiat_p224_addcarryx_u32(&x194, &x195, x193, x178, x150);
  uint32_t x196;
  uint32_t x197;
  fiat_p224_mulx_u32(&x196, &x197, x180, UINT32_C(0xffffffff));
  uint32_t x198;
  uint32_t x199;
  fiat_p224_mulx_u32(&x198, &x199, x196, UINT32_C(0xffffffff));
  uint32_t x200;
  uint32_t x201;
  fiat_p224_mulx_u32(&x200, &x201, x196, UINT32_C(0xffffffff));
  uint32_t x202;
  uint32_t x203;
  fiat_p224_mulx_u32(&x202, &x203, x196, UINT32_C(0xffffffff));
  uint32_t x204;
  uint32_t x205;
  fiat_p224_mulx_u32(&x204, &x205, x196, UINT32_C(0xffffffff));
  uint32_t x206;
  fiat_p224_uint1 x207;
  fiat_p224_addcarryx_u32(&x206, &x207, 0x0, x202, x205);
  uint32_t x208;
  fiat_p224_uint1 x209;
  fiat_p224_addcarryx_u32(&x208, &x209, x207, x200, x203);
  uint32_t x210;
  fiat_p224_uint1 x211;
  fiat_p224_addcarryx_u32(&x210, &x211, x209, x198, x201);
  uint32_t x212;
  fiat_p224_uint1 x213;
  fiat_p224_addcarryx_u32(&x212, &x213, x211, 0x0, x199);
  uint32_t x214;
  fiat_p224_uint1 x215;
  fiat_p224_addcarryx_u32(&x214, &x215, 0x0, x196, x180);
  uint32_t x216;
  fiat_p224_uint1 x217;
  fiat_p224_addcarryx_u32(&x216, &x217, x215, 0x0, x182);
  uint32_t x218;
  fiat_p224_uint1 x219;
  fiat_p224_addcarryx_u32(&x218, &x219, x217, 0x0, x184);
  uint32_t x220;
  fiat_p224_uint1 x221;
  fiat_p224_addcarryx_u32(&x220, &x221, x219, x204, x186);
  uint32_t x222;
  fiat_p224_uint1 x223;
  fiat_p224_addcarryx_u32(&x222, &x223, x221, x206, x188);
  uint32_t x224;
  fiat_p224_uint1 x225;
  fiat_p224_addcarryx_u32(&x224, &x225, x223, x208, x190);
  uint32_t x226;
  fiat_p224_uint1 x227;
  fiat_p224_addcarryx_u32(&x226, &x227, x225, x210, x192);
  uint32_t x228;
  fiat_p224_uint1 x229;
  fiat_p224_addcarryx_u32(&x228, &x229, x227, x212, x194);
  uint32_t x230;
  fiat_p224_uint1 x231;
  fiat_p224_addcarryx_u32(&x230, &x231, x229, 0x0, x195);
  uint32_t x232;
  uint32_t x233;
  fiat_p224_mulx_u32(&x232, &x233, x3, (arg1[6]));
  uint32_t x234;
  uint32_t x235;
  fiat_p224_mulx_u32(&x234, &x235, x3, (arg1[5]));
  uint32_t x236;
  uint32_t x237;
  fiat_p224_mulx_u32(&x236, &x237, x3, (arg1[4]));
  uint32_t x238;
  uint32_t x239;
  fiat_p224_mulx_u32(&x238, &x239, x3, (arg1[3]));
  uint32_t x240;
  uint32_t x241;
  fiat_p224_mulx_u32(&x240, &x241, x3, (arg1[2]));
  uint32_t x242;
  uint32_t x243;
  fiat_p224_mulx_u32(&x242, &x243, x3, (arg1[1]));
  uint32_t x244;
  uint32_t x245;
  fiat_p224_mulx_u32(&x244, &x245, x3, (arg1[0]));
  uint32_t x246;
  fiat_p224_uint1 x247;
  fiat_p224_addcarryx_u32(&x246, &x247, 0x0, x242, x245);
  uint32_t x248;
  fiat_p224_uint1 x249;
  fiat_p224_addcarryx_u32(&x248, &x249, x247, x240, x243);
  uint32_t x250;
  fiat_p224_uint1 x251;
  fiat_p224_addcarryx_u32(&x250, &x251, x249, x238, x241);
  uint32_t x252;
  fiat_p224_uint1 x253;
  fiat_p224_addcarryx_u32(&x252, &x253, x251, x236, x239);
  uint32_t x254;
  fiat_p224_uint1 x255;
  fiat_p224_addcarryx_u32(&x254, &x255, x253, x234, x237);
  uint32_t x256;
  fiat_p224_uint1 x257;
  fiat_p224_addcarryx_u32(&x256, &x257, x255, x232, x235);
  uint32_t x258;
  fiat_p224_uint1 x259;
  fiat_p224_addcarryx_u32(&x258, &x259, x257, 0x0, x233);
  uint32_t x260;
  fiat_p224_uint1 x261;
  fiat_p224_addcarryx_u32(&x260, &x261, 0x0, x244, x216);
  uint32_t x262;
  fiat_p224_uint1 x263;
  fiat_p224_addcarryx_u32(&x262, &x263, x261, x246, x218);
  uint32_t x264;
  fiat_p224_uint1 x265;
  fiat_p224_addcarryx_u32(&x264, &x265, x263, x248, x220);
  uint32_t x266;
  fiat_p224_uint1 x267;
  fiat_p224_addcarryx_u32(&x266, &x267, x265, x250, x222);
  uint32_t x268;
  fiat_p224_uint1 x269;
  fiat_p224_addcarryx_u32(&x268, &x269, x267, x252, x224);
  uint32_t x270;
  fiat_p224_uint1 x271;
  fiat_p224_addcarryx_u32(&x270, &x271, x269, x254, x226);
  uint32_t x272;
  fiat_p224_uint1 x273;
  fiat_p224_addcarryx_u32(&x272, &x273, x271, x256, x228);
  uint32_t x274;
  fiat_p224_uint1 x275;
  fiat_p224_addcarryx_u32(&x274, &x275, x273, x258, x230);
  uint32_t x276;
  uint32_t x277;
  fiat_p224_mulx_u32(&x276, &x277, x260, UINT32_C(0xffffffff));
  uint32_t x278;
  uint32_t x279;
  fiat_p224_mulx_u32(&x278, &x279, x276, UINT32_C(0xffffffff));
  uint32_t x280;
  uint32_t x281;
  fiat_p224_mulx_u32(&x280, &x281, x276, UINT32_C(0xffffffff));
  uint32_t x282;
  uint32_t x283;
  fiat_p224_mulx_u32(&x282, &x283, x276, UINT32_C(0xffffffff));
  uint32_t x284;
  uint32_t x285;
  fiat_p224_mulx_u32(&x284, &x285, x276, UINT32_C(0xffffffff));
  uint32_t x286;
  fiat_p224_uint1 x287;
  fiat_p224_addcarryx_u32(&x286, &x287, 0x0, x282, x285);
  uint32_t x288;
  fiat_p224_uint1 x289;
  fiat_p224_addcarryx_u32(&x288, &x289, x287, x280, x283);
  uint32_t x290;
  fiat_p224_uint1 x291;
  fiat_p224_addcarryx_u32(&x290, &x291, x289, x278, x281);
  uint32_t x292;
  fiat_p224_uint1 x293;
  fiat_p224_addcarryx_u32(&x292, &x293, x291, 0x0, x279);
  uint32_t x294;
  fiat_p224_uint1 x295;
  fiat_p224_addcarryx_u32(&x294, &x295, 0x0, x276, x260);
  uint32_t x296;
  fiat_p224_uint1 x297;
  fiat_p224_addcarryx_u32(&x296, &x297, x295, 0x0, x262);
  uint32_t x298;
  fiat_p224_uint1 x299;
  fiat_p224_addcarryx_u32(&x298, &x299, x297, 0x0, x264);
  uint32_t x300;
  fiat_p224_uint1 x301;
  fiat_p224_addcarryx_u32(&x300, &x301, x299, x284, x266);
  uint32_t x302;
  fiat_p224_uint1 x303;
  fiat_p224_addcarryx_u32(&x302, &x303, x301, x286, x268);
  uint32_t x304;
  fiat_p224_uint1 x305;
  fiat_p224_addcarryx_u32(&x304, &x305, x303, x288, x270);
  uint32_t x306;
  fiat_p224_uint1 x307;
  fiat_p224_addcarryx_u32(&x306, &x307, x305, x290, x272);
  uint32_t x308;
  fiat_p224_uint1 x309;
  fiat_p224_addcarryx_u32(&x308, &x309, x307, x292, x274);
  uint32_t x310;
  fiat_p224_uint1 x311;
  fiat_p224_addcarryx_u32(&x310, &x311, x309, 0x0, x275);
  uint32_t x312;
  uint32_t x313;
  fiat_p224_mulx_u32(&x312, &x313, x4, (arg1[6]));
  uint32_t x314;
  uint32_t x315;
  fiat_p224_mulx_u32(&x314, &x315, x4, (arg1[5]));
  uint32_t x316;
  uint32_t x317;
  fiat_p224_mulx_u32(&x316, &x317, x4, (arg1[4]));
  uint32_t x318;
  uint32_t x319;
  fiat_p224_mulx_u32(&x318, &x319, x4, (arg1[3]));
  uint32_t x320;
  uint32_t x321;
  fiat_p224_mulx_u32(&x320, &x321, x4, (arg1[2]));
  uint32_t x322;
  uint32_t x323;
  fiat_p224_mulx_u32(&x322, &x323, x4, (arg1[1]));
  uint32_t x324;
  uint32_t x325;
  fiat_p224_mulx_u32(&x324, &x325, x4, (arg1[0]));
  uint32_t x326;
  fiat_p224_uint1 x327;
  fiat_p224_addcarryx_u32(&x326, &x327, 0x0, x322, x325);
  uint32_t x328;
  fiat_p224_uint1 x329;
  fiat_p224_addcarryx_u32(&x328, &x329, x327, x320, x323);
  uint32_t x330;
  fiat_p224_uint1 x331;
  fiat_p224_addcarryx_u32(&x330, &x331, x329, x318, x321);
  uint32_t x332;
  fiat_p224_uint1 x333;
  fiat_p224_addcarryx_u32(&x332, &x333, x331, x316, x319);
  uint32_t x334;
  fiat_p224_uint1 x335;
  fiat_p224_addcarryx_u32(&x334, &x335, x333, x314, x317);
  uint32_t x336;
  fiat_p224_uint1 x337;
  fiat_p224_addcarryx_u32(&x336, &x337, x335, x312, x315);
  uint32_t x338;
  fiat_p224_uint1 x339;
  fiat_p224_addcarryx_u32(&x338, &x339, x337, 0x0, x313);
  uint32_t x340;
  fiat_p224_uint1 x341;
  fiat_p224_addcarryx_u32(&x340, &x341, 0x0, x324, x296);
  uint32_t x342;
  fiat_p224_uint1 x343;
  fiat_p224_addcarryx_u32(&x342, &x343, x341, x326, x298);
  uint32_t x344;
  fiat_p224_uint1 x345;
  fiat_p224_addcarryx_u32(&x344, &x345, x343, x328, x300);
  uint32_t x346;
  fiat_p224_uint1 x347;
  fiat_p224_addcarryx_u32(&x346, &x347, x345, x330, x302);
  uint32_t x348;
  fiat_p224_uint1 x349;
  fiat_p224_addcarryx_u32(&x348, &x349, x347, x332, x304);
  uint32_t x350;
  fiat_p224_uint1 x351;
  fiat_p224_addcarryx_u32(&x350, &x351, x349, x334, x306);
  uint32_t x352;
  fiat_p224_uint1 x353;
  fiat_p224_addcarryx_u32(&x352, &x353, x351, x336, x308);
  uint32_t x354;
  fiat_p224_uint1 x355;
  fiat_p224_addcarryx_u32(&x354, &x355, x353, x338, x310);
  uint32_t x356;
  uint32_t x357;
  fiat_p224_mulx_u32(&x356, &x357, x340, UINT32_C(0xffffffff));
  uint32_t x358;
  uint32_t x359;
  fiat_p224_mulx_u32(&x358, &x359, x356, UINT32_C(0xffffffff));
  uint32_t x360;
  uint32_t x361;
  fiat_p224_mulx_u32(&x360, &x361, x356, UINT32_C(0xffffffff));
  uint32_t x362;
  uint32_t x363;
  fiat_p224_mulx_u32(&x362, &x363, x356, UINT32_C(0xffffffff));
  uint32_t x364;
  uint32_t x365;
  fiat_p224_mulx_u32(&x364, &x365, x356, UINT32_C(0xffffffff));
  uint32_t x366;
  fiat_p224_uint1 x367;
  fiat_p224_addcarryx_u32(&x366, &x367, 0x0, x362, x365);
  uint32_t x368;
  fiat_p224_uint1 x369;
  fiat_p224_addcarryx_u32(&x368, &x369, x367, x360, x363);
  uint32_t x370;
  fiat_p224_uint1 x371;
  fiat_p224_addcarryx_u32(&x370, &x371, x369, x358, x361);
  uint32_t x372;
  fiat_p224_uint1 x373;
  fiat_p224_addcarryx_u32(&x372, &x373, x371, 0x0, x359);
  uint32_t x374;
  fiat_p224_uint1 x375;
  fiat_p224_addcarryx_u32(&x374, &x375, 0x0, x356, x340);
  uint32_t x376;
  fiat_p224_uint1 x377;
  fiat_p224_addcarryx_u32(&x376, &x377, x375, 0x0, x342);
  uint32_t x378;
  fiat_p224_uint1 x379;
  fiat_p224_addcarryx_u32(&x378, &x379, x377, 0x0, x344);
  uint32_t x380;
  fiat_p224_uint1 x381;
  fiat_p224_addcarryx_u32(&x380, &x381, x379, x364, x346);
  uint32_t x382;
  fiat_p224_uint1 x383;
  fiat_p224_addcarryx_u32(&x382, &x383, x381, x366, x348);
  uint32_t x384;
  fiat_p224_uint1 x385;
  fiat_p224_addcarryx_u32(&x384, &x385, x383, x368, x350);
  uint32_t x386;
  fiat_p224_uint1 x387;
  fiat_p224_addcarryx_u32(&x386, &x387, x385, x370, x352);
  uint32_t x388;
  fiat_p224_uint1 x389;
  fiat_p224_addcarryx_u32(&x388, &x389, x387, x372, x354);
  uint32_t x390;
  fiat_p224_uint1 x391;
  fiat_p224_addcarryx_u32(&x390, &x391, x389, 0x0, x355);
  uint32_t x392;
  uint32_t x393;
  fiat_p224_mulx_u32(&x392, &x393, x5, (arg1[6]));
  uint32_t x394;
  uint32_t x395;
  fiat_p224_mulx_u32(&x394, &x395, x5, (arg1[5]));
  uint32_t x396;
  uint32_t x397;
  fiat_p224_mulx_u32(&x396, &x397, x5, (arg1[4]));
  uint32_t x398;
  uint32_t x399;
  fiat_p224_mulx_u32(&x398, &x399, x5, (arg1[3]));
  uint32_t x400;
  uint32_t x401;
  fiat_p224_mulx_u32(&x400, &x401, x5, (arg1[2]));
  uint32_t x402;
  uint32_t x403;
  fiat_p224_mulx_u32(&x402, &x403, x5, (arg1[1]));
  uint32_t x404;
  uint32_t x405;
  fiat_p224_mulx_u32(&x404, &x405, x5, (arg1[0]));
  uint32_t x406;
  fiat_p224_uint1 x407;
  fiat_p224_addcarryx_u32(&x406, &x407, 0x0, x402, x405);
  uint32_t x408;
  fiat_p224_uint1 x409;
  fiat_p224_addcarryx_u32(&x408, &x409, x407, x400, x403);
  uint32_t x410;
  fiat_p224_uint1 x411;
  fiat_p224_addcarryx_u32(&x410, &x411, x409, x398, x401);
  uint32_t x412;
  fiat_p224_uint1 x413;
  fiat_p224_addcarryx_u32(&x412, &x413, x411, x396, x399);
  uint32_t x414;
  fiat_p224_uint1 x415;
  fiat_p224_addcarryx_u32(&x414, &x415, x413, x394, x397);
  uint32_t x416;
  fiat_p224_uint1 x417;
  fiat_p224_addcarryx_u32(&x416, &x417, x415, x392, x395);
  uint32_t x418;
  fiat_p224_uint1 x419;
  fiat_p224_addcarryx_u32(&x418, &x419, x417, 0x0, x393);
  uint32_t x420;
  fiat_p224_uint1 x421;
  fiat_p224_addcarryx_u32(&x420, &x421, 0x0, x404, x376);
  uint32_t x422;
  fiat_p224_uint1 x423;
  fiat_p224_addcarryx_u32(&x422, &x423, x421, x406, x378);
  uint32_t x424;
  fiat_p224_uint1 x425;
  fiat_p224_addcarryx_u32(&x424, &x425, x423, x408, x380);
  uint32_t x426;
  fiat_p224_uint1 x427;
  fiat_p224_addcarryx_u32(&x426, &x427, x425, x410, x382);
  uint32_t x428;
  fiat_p224_uint1 x429;
  fiat_p224_addcarryx_u32(&x428, &x429, x427, x412, x384);
  uint32_t x430;
  fiat_p224_uint1 x431;
  fiat_p224_addcarryx_u32(&x430, &x431, x429, x414, x386);
  uint32_t x432;
  fiat_p224_uint1 x433;
  fiat_p224_addcarryx_u32(&x432, &x433, x431, x416, x388);
  uint32_t x434;
  fiat_p224_uint1 x435;
  fiat_p224_addcarryx_u32(&x434, &x435, x433, x418, x390);
  uint32_t x436;
  uint32_t x437;
  fiat_p224_mulx_u32(&x436, &x437, x420, UINT32_C(0xffffffff));
  uint32_t x438;
  uint32_t x439;
  fiat_p224_mulx_u32(&x438, &x439, x436, UINT32_C(0xffffffff));
  uint32_t x440;
  uint32_t x441;
  fiat_p224_mulx_u32(&x440, &x441, x436, UINT32_C(0xffffffff));
  uint32_t x442;
  uint32_t x443;
  fiat_p224_mulx_u32(&x442, &x443, x436, UINT32_C(0xffffffff));
  uint32_t x444;
  uint32_t x445;
  fiat_p224_mulx_u32(&x444, &x445, x436, UINT32_C(0xffffffff));
  uint32_t x446;
  fiat_p224_uint1 x447;
  fiat_p224_addcarryx_u32(&x446, &x447, 0x0, x442, x445);
  uint32_t x448;
  fiat_p224_uint1 x449;
  fiat_p224_addcarryx_u32(&x448, &x449, x447, x440, x443);
  uint32_t x450;
  fiat_p224_uint1 x451;
  fiat_p224_addcarryx_u32(&x450, &x451, x449, x438, x441);
  uint32_t x452;
  fiat_p224_uint1 x453;
  fiat_p224_addcarryx_u32(&x452, &x453, x451, 0x0, x439);
  uint32_t x454;
  fiat_p224_uint1 x455;
  fiat_p224_addcarryx_u32(&x454, &x455, 0x0, x436, x420);
  uint32_t x456;
  fiat_p224_uint1 x457;
  fiat_p224_addcarryx_u32(&x456, &x457, x455, 0x0, x422);
  uint32_t x458;
  fiat_p224_uint1 x459;
  fiat_p224_addcarryx_u32(&x458, &x459, x457, 0x0, x424);
  uint32_t x460;
  fiat_p224_uint1 x461;
  fiat_p224_addcarryx_u32(&x460, &x461, x459, x444, x426);
  uint32_t x462;
  fiat_p224_uint1 x463;
  fiat_p224_addcarryx_u32(&x462, &x463, x461, x446, x428);
  uint32_t x464;
  fiat_p224_uint1 x465;
  fiat_p224_addcarryx_u32(&x464, &x465, x463, x448, x430);
  uint32_t x466;
  fiat_p224_uint1 x467;
  fiat_p224_addcarryx_u32(&x466, &x467, x465, x450, x432);
  uint32_t x468;
  fiat_p224_uint1 x469;
  fiat_p224_addcarryx_u32(&x468, &x469, x467, x452, x434);
  uint32_t x470;
  fiat_p224_uint1 x471;
  fiat_p224_addcarryx_u32(&x470, &x471, x469, 0x0, x435);
  uint32_t x472;
  uint32_t x473;
  fiat_p224_mulx_u32(&x472, &x473, x6, (arg1[6]));
  uint32_t x474;
  uint32_t x475;
  fiat_p224_mulx_u32(&x474, &x475, x6, (arg1[5]));
  uint32_t x476;
  uint32_t x477;
  fiat_p224_mulx_u32(&x476, &x477, x6, (arg1[4]));
  uint32_t x478;
  uint32_t x479;
  fiat_p224_mulx_u32(&x478, &x479, x6, (arg1[3]));
  uint32_t x480;
  uint32_t x481;
  fiat_p224_mulx_u32(&x480, &x481, x6, (arg1[2]));
  uint32_t x482;
  uint32_t x483;
  fiat_p224_mulx_u32(&x482, &x483, x6, (arg1[1]));
  uint32_t x484;
  uint32_t x485;
  fiat_p224_mulx_u32(&x484, &x485, x6, (arg1[0]));
  uint32_t x486;
  fiat_p224_uint1 x487;
  fiat_p224_addcarryx_u32(&x486, &x487, 0x0, x482, x485);
  uint32_t x488;
  fiat_p224_uint1 x489;
  fiat_p224_addcarryx_u32(&x488, &x489, x487, x480, x483);
  uint32_t x490;
  fiat_p224_uint1 x491;
  fiat_p224_addcarryx_u32(&x490, &x491, x489, x478, x481);
  uint32_t x492;
  fiat_p224_uint1 x493;
  fiat_p224_addcarryx_u32(&x492, &x493, x491, x476, x479);
  uint32_t x494;
  fiat_p224_uint1 x495;
  fiat_p224_addcarryx_u32(&x494, &x495, x493, x474, x477);
  uint32_t x496;
  fiat_p224_uint1 x497;
  fiat_p224_addcarryx_u32(&x496, &x497, x495, x472, x475);
  uint32_t x498;
  fiat_p224_uint1 x499;
  fiat_p224_addcarryx_u32(&x498, &x499, x497, 0x0, x473);
  uint32_t x500;
  fiat_p224_uint1 x501;
  fiat_p224_addcarryx_u32(&x500, &x501, 0x0, x484, x456);
  uint32_t x502;
  fiat_p224_uint1 x503;
  fiat_p224_addcarryx_u32(&x502, &x503, x501, x486, x458);
  uint32_t x504;
  fiat_p224_uint1 x505;
  fiat_p224_addcarryx_u32(&x504, &x505, x503, x488, x460);
  uint32_t x506;
  fiat_p224_uint1 x507;
  fiat_p224_addcarryx_u32(&x506, &x507, x505, x490, x462);
  uint32_t x508;
  fiat_p224_uint1 x509;
  fiat_p224_addcarryx_u32(&x508, &x509, x507, x492, x464);
  uint32_t x510;
  fiat_p224_uint1 x511;
  fiat_p224_addcarryx_u32(&x510, &x511, x509, x494, x466);
  uint32_t x512;
  fiat_p224_uint1 x513;
  fiat_p224_addcarryx_u32(&x512, &x513, x511, x496, x468);
  uint32_t x514;
  fiat_p224_uint1 x515;
  fiat_p224_addcarryx_u32(&x514, &x515, x513, x498, x470);
  uint32_t x516;
  uint32_t x517;
  fiat_p224_mulx_u32(&x516, &x517, x500, UINT32_C(0xffffffff));
  uint32_t x518;
  uint32_t x519;
  fiat_p224_mulx_u32(&x518, &x519, x516, UINT32_C(0xffffffff));
  uint32_t x520;
  uint32_t x521;
  fiat_p224_mulx_u32(&x520, &x521, x516, UINT32_C(0xffffffff));
  uint32_t x522;
  uint32_t x523;
  fiat_p224_mulx_u32(&x522, &x523, x516, UINT32_C(0xffffffff));
  uint32_t x524;
  uint32_t x525;
  fiat_p224_mulx_u32(&x524, &x525, x516, UINT32_C(0xffffffff));
  uint32_t x526;
  fiat_p224_uint1 x527;
  fiat_p224_addcarryx_u32(&x526, &x527, 0x0, x522, x525);
  uint32_t x528;
  fiat_p224_uint1 x529;
  fiat_p224_addcarryx_u32(&x528, &x529, x527, x520, x523);
  uint32_t x530;
  fiat_p224_uint1 x531;
  fiat_p224_addcarryx_u32(&x530, &x531, x529, x518, x521);
  uint32_t x532;
  fiat_p224_uint1 x533;
  fiat_p224_addcarryx_u32(&x532, &x533, x531, 0x0, x519);
  uint32_t x534;
  fiat_p224_uint1 x535;
  fiat_p224_addcarryx_u32(&x534, &x535, 0x0, x516, x500);
  uint32_t x536;
  fiat_p224_uint1 x537;
  fiat_p224_addcarryx_u32(&x536, &x537, x535, 0x0, x502);
  uint32_t x538;
  fiat_p224_uint1 x539;
  fiat_p224_addcarryx_u32(&x538, &x539, x537, 0x0, x504);
  uint32_t x540;
  fiat_p224_uint1 x541;
  fiat_p224_addcarryx_u32(&x540, &x541, x539, x524, x506);
  uint32_t x542;
  fiat_p224_uint1 x543;
  fiat_p224_addcarryx_u32(&x542, &x543, x541, x526, x508);
  uint32_t x544;
  fiat_p224_uint1 x545;
  fiat_p224_addcarryx_u32(&x544, &x545, x543, x528, x510);
  uint32_t x546;
  fiat_p224_uint1 x547;
  fiat_p224_addcarryx_u32(&x546, &x547, x545, x530, x512);
  uint32_t x548;
  fiat_p224_uint1 x549;
  fiat_p224_addcarryx_u32(&x548, &x549, x547, x532, x514);
  uint32_t x550;
  fiat_p224_uint1 x551;
  fiat_p224_addcarryx_u32(&x550, &x551, x549, 0x0, x515);
  uint32_t x552;
  fiat_p224_uint1 x553;
  fiat_p224_subborrowx_u32(&x552, &x553, 0x0, x536, 0x1);
  uint32_t x554;
  fiat_p224_uint1 x555;
  fiat_p224_subborrowx_u32(&x554, &x555, x553, x538, 0x0);
  uint32_t x556;
  fiat_p224_uint1 x557;
  fiat_p224_subborrowx_u32(&x556, &x557, x555, x540, 0x0);
  uint32_t x558;
  fiat_p224_uint1 x559;
  fiat_p224_subborrowx_u32(&x558, &x559, x557, x542, UINT32_C(0xffffffff));
  uint32_t x560;
  fiat_p224_uint1 x561;
  fiat_p224_subborrowx_u32(&x560, &x561, x559, x544, UINT32_C(0xffffffff));
  uint32_t x562;
  fiat_p224_uint1 x563;
  fiat_p224_subborrowx_u32(&x562, &x563, x561, x546, UINT32_C(0xffffffff));
  uint32_t x564;
  fiat_p224_uint1 x565;
  fiat_p224_subborrowx_u32(&x564, &x565, x563, x548, UINT32_C(0xffffffff));
  uint32_t x566;
  fiat_p224_uint1 x567;
  fiat_p224_subborrowx_u32(&x566, &x567, x565, x550, 0x0);
  uint32_t x568;
  fiat_p224_cmovznz_u32(&x568, x567, x552, x536);
  uint32_t x569;
  fiat_p224_cmovznz_u32(&x569, x567, x554, x538);
  uint32_t x570;
  fiat_p224_cmovznz_u32(&x570, x567, x556, x540);
  uint32_t x571;
  fiat_p224_cmovznz_u32(&x571, x567, x558, x542);
  uint32_t x572;
  fiat_p224_cmovznz_u32(&x572, x567, x560, x544);
  uint32_t x573;
  fiat_p224_cmovznz_u32(&x573, x567, x562, x546);
  uint32_t x574;
  fiat_p224_cmovznz_u32(&x574, x567, x564, x548);
  out1[0] = x568;
  out1[1] = x569;
  out1[2] = x570;
  out1[3] = x571;
  out1[4] = x572;
  out1[5] = x573;
  out1[6] = x574;
}

/*
 * Input Bounds:
 *   arg1: [[0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff]]
 *   arg2: [[0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff]]
 * Output Bounds:
 *   out1: [[0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff]]
 */
static void fiat_p224_add(uint32_t out1[7], const uint32_t arg1[7], const uint32_t arg2[7]) {
  uint32_t x1;
  fiat_p224_uint1 x2;
  fiat_p224_addcarryx_u32(&x1, &x2, 0x0, (arg2[0]), (arg1[0]));
  uint32_t x3;
  fiat_p224_uint1 x4;
  fiat_p224_addcarryx_u32(&x3, &x4, x2, (arg2[1]), (arg1[1]));
  uint32_t x5;
  fiat_p224_uint1 x6;
  fiat_p224_addcarryx_u32(&x5, &x6, x4, (arg2[2]), (arg1[2]));
  uint32_t x7;
  fiat_p224_uint1 x8;
  fiat_p224_addcarryx_u32(&x7, &x8, x6, (arg2[3]), (arg1[3]));
  uint32_t x9;
  fiat_p224_uint1 x10;
  fiat_p224_addcarryx_u32(&x9, &x10, x8, (arg2[4]), (arg1[4]));
  uint32_t x11;
  fiat_p224_uint1 x12;
  fiat_p224_addcarryx_u32(&x11, &x12, x10, (arg2[5]), (arg1[5]));
  uint32_t x13;
  fiat_p224_uint1 x14;
  fiat_p224_addcarryx_u32(&x13, &x14, x12, (arg2[6]), (arg1[6]));
  uint32_t x15;
  fiat_p224_uint1 x16;
  fiat_p224_subborrowx_u32(&x15, &x16, 0x0, x1, 0x1);
  uint32_t x17;
  fiat_p224_uint1 x18;
  fiat_p224_subborrowx_u32(&x17, &x18, x16, x3, 0x0);
  uint32_t x19;
  fiat_p224_uint1 x20;
  fiat_p224_subborrowx_u32(&x19, &x20, x18, x5, 0x0);
  uint32_t x21;
  fiat_p224_uint1 x22;
  fiat_p224_subborrowx_u32(&x21, &x22, x20, x7, UINT32_C(0xffffffff));
  uint32_t x23;
  fiat_p224_uint1 x24;
  fiat_p224_subborrowx_u32(&x23, &x24, x22, x9, UINT32_C(0xffffffff));
  uint32_t x25;
  fiat_p224_uint1 x26;
  fiat_p224_subborrowx_u32(&x25, &x26, x24, x11, UINT32_C(0xffffffff));
  uint32_t x27;
  fiat_p224_uint1 x28;
  fiat_p224_subborrowx_u32(&x27, &x28, x26, x13, UINT32_C(0xffffffff));
  uint32_t x29;
  fiat_p224_uint1 x30;
  fiat_p224_subborrowx_u32(&x29, &x30, x28, x14, 0x0);
  uint32_t x31;
  fiat_p224_cmovznz_u32(&x31, x30, x15, x1);
  uint32_t x32;
  fiat_p224_cmovznz_u32(&x32, x30, x17, x3);
  uint32_t x33;
  fiat_p224_cmovznz_u32(&x33, x30, x19, x5);
  uint32_t x34;
  fiat_p224_cmovznz_u32(&x34, x30, x21, x7);
  uint32_t x35;
  fiat_p224_cmovznz_u32(&x35, x30, x23, x9);
  uint32_t x36;
  fiat_p224_cmovznz_u32(&x36, x30, x25, x11);
  uint32_t x37;
  fiat_p224_cmovznz_u32(&x37, x30, x27, x13);
  out1[0] = x31;
  out1[1] = x32;
  out1[2] = x33;
  out1[3] = x34;
  out1[4] = x35;
  out1[5] = x36;
  out1[6] = x37;
}

/*
 * Input Bounds:
 *   arg1: [[0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff]]
 *   arg2: [[0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff]]
 * Output Bounds:
 *   out1: [[0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff]]
 */
static void fiat_p224_sub(uint32_t out1[7], const uint32_t arg1[7], const uint32_t arg2[7]) {
  uint32_t x1;
  fiat_p224_uint1 x2;
  fiat_p224_subborrowx_u32(&x1, &x2, 0x0, (arg1[0]), (arg2[0]));
  uint32_t x3;
  fiat_p224_uint1 x4;
  fiat_p224_subborrowx_u32(&x3, &x4, x2, (arg1[1]), (arg2[1]));
  uint32_t x5;
  fiat_p224_uint1 x6;
  fiat_p224_subborrowx_u32(&x5, &x6, x4, (arg1[2]), (arg2[2]));
  uint32_t x7;
  fiat_p224_uint1 x8;
  fiat_p224_subborrowx_u32(&x7, &x8, x6, (arg1[3]), (arg2[3]));
  uint32_t x9;
  fiat_p224_uint1 x10;
  fiat_p224_subborrowx_u32(&x9, &x10, x8, (arg1[4]), (arg2[4]));
  uint32_t x11;
  fiat_p224_uint1 x12;
  fiat_p224_subborrowx_u32(&x11, &x12, x10, (arg1[5]), (arg2[5]));
  uint32_t x13;
  fiat_p224_uint1 x14;
  fiat_p224_subborrowx_u32(&x13, &x14, x12, (arg1[6]), (arg2[6]));
  uint32_t x15;
  fiat_p224_cmovznz_u32(&x15, x14, 0x0, UINT32_C(0xffffffff));
  uint32_t x16;
  fiat_p224_uint1 x17;
  fiat_p224_addcarryx_u32(&x16, &x17, 0x0, (fiat_p224_uint1)(x15 & 0x1), x1);
  uint32_t x18;
  fiat_p224_uint1 x19;
  fiat_p224_addcarryx_u32(&x18, &x19, x17, 0x0, x3);
  uint32_t x20;
  fiat_p224_uint1 x21;
  fiat_p224_addcarryx_u32(&x20, &x21, x19, 0x0, x5);
  uint32_t x22;
  fiat_p224_uint1 x23;
  fiat_p224_addcarryx_u32(&x22, &x23, x21, (x15 & UINT32_C(0xffffffff)), x7);
  uint32_t x24;
  fiat_p224_uint1 x25;
  fiat_p224_addcarryx_u32(&x24, &x25, x23, (x15 & UINT32_C(0xffffffff)), x9);
  uint32_t x26;
  fiat_p224_uint1 x27;
  fiat_p224_addcarryx_u32(&x26, &x27, x25, (x15 & UINT32_C(0xffffffff)), x11);
  uint32_t x28;
  fiat_p224_uint1 x29;
  fiat_p224_addcarryx_u32(&x28, &x29, x27, (x15 & UINT32_C(0xffffffff)), x13);
  out1[0] = x16;
  out1[1] = x18;
  out1[2] = x20;
  out1[3] = x22;
  out1[4] = x24;
  out1[5] = x26;
  out1[6] = x28;
}

/*
 * Input Bounds:
 *   arg1: [[0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff]]
 * Output Bounds:
 *   out1: [[0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff]]
 */
static void fiat_p224_opp(uint32_t out1[7], const uint32_t arg1[7]) {
  uint32_t x1;
  fiat_p224_uint1 x2;
  fiat_p224_subborrowx_u32(&x1, &x2, 0x0, 0x0, (arg1[0]));
  uint32_t x3;
  fiat_p224_uint1 x4;
  fiat_p224_subborrowx_u32(&x3, &x4, x2, 0x0, (arg1[1]));
  uint32_t x5;
  fiat_p224_uint1 x6;
  fiat_p224_subborrowx_u32(&x5, &x6, x4, 0x0, (arg1[2]));
  uint32_t x7;
  fiat_p224_uint1 x8;
  fiat_p224_subborrowx_u32(&x7, &x8, x6, 0x0, (arg1[3]));
  uint32_t x9;
  fiat_p224_uint1 x10;
  fiat_p224_subborrowx_u32(&x9, &x10, x8, 0x0, (arg1[4]));
  uint32_t x11;
  fiat_p224_uint1 x12;
  fiat_p224_subborrowx_u32(&x11, &x12, x10, 0x0, (arg1[5]));
  uint32_t x13;
  fiat_p224_uint1 x14;
  fiat_p224_subborrowx_u32(&x13, &x14, x12, 0x0, (arg1[6]));
  uint32_t x15;
  fiat_p224_cmovznz_u32(&x15, x14, 0x0, UINT32_C(0xffffffff));
  uint32_t x16;
  fiat_p224_uint1 x17;
  fiat_p224_addcarryx_u32(&x16, &x17, 0x0, (fiat_p224_uint1)(x15 & 0x1), x1);
  uint32_t x18;
  fiat_p224_uint1 x19;
  fiat_p224_addcarryx_u32(&x18, &x19, x17, 0x0, x3);
  uint32_t x20;
  fiat_p224_uint1 x21;
  fiat_p224_addcarryx_u32(&x20, &x21, x19, 0x0, x5);
  uint32_t x22;
  fiat_p224_uint1 x23;
  fiat_p224_addcarryx_u32(&x22, &x23, x21, (x15 & UINT32_C(0xffffffff)), x7);
  uint32_t x24;
  fiat_p224_uint1 x25;
  fiat_p224_addcarryx_u32(&x24, &x25, x23, (x15 & UINT32_C(0xffffffff)), x9);
  uint32_t x26;
  fiat_p224_uint1 x27;
  fiat_p224_addcarryx_u32(&x26, &x27, x25, (x15 & UINT32_C(0xffffffff)), x11);
  uint32_t x28;
  fiat_p224_uint1 x29;
  fiat_p224_addcarryx_u32(&x28, &x29, x27, (x15 & UINT32_C(0xffffffff)), x13);
  out1[0] = x16;
  out1[1] = x18;
  out1[2] = x20;
  out1[3] = x22;
  out1[4] = x24;
  out1[5] = x26;
  out1[6] = x28;
}

/*
 * Input Bounds:
 *   arg1: [[0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff]]
 * Output Bounds:
 *   out1: [[0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff]]
 */
static void fiat_p224_from_montgomery(uint32_t out1[7], const uint32_t arg1[7]) {
  uint32_t x1 = (arg1[0]);
  uint32_t x2;
  uint32_t x3;
  fiat_p224_mulx_u32(&x2, &x3, x1, UINT32_C(0xffffffff));
  uint32_t x4;
  uint32_t x5;
  fiat_p224_mulx_u32(&x4, &x5, x2, UINT32_C(0xffffffff));
  uint32_t x6;
  uint32_t x7;
  fiat_p224_mulx_u32(&x6, &x7, x2, UINT32_C(0xffffffff));
  uint32_t x8;
  uint32_t x9;
  fiat_p224_mulx_u32(&x8, &x9, x2, UINT32_C(0xffffffff));
  uint32_t x10;
  uint32_t x11;
  fiat_p224_mulx_u32(&x10, &x11, x2, UINT32_C(0xffffffff));
  uint32_t x12;
  fiat_p224_uint1 x13;
  fiat_p224_addcarryx_u32(&x12, &x13, 0x0, x8, x11);
  uint32_t x14;
  fiat_p224_uint1 x15;
  fiat_p224_addcarryx_u32(&x14, &x15, x13, x6, x9);
  uint32_t x16;
  fiat_p224_uint1 x17;
  fiat_p224_addcarryx_u32(&x16, &x17, x15, x4, x7);
  uint32_t x18;
  fiat_p224_uint1 x19;
  fiat_p224_addcarryx_u32(&x18, &x19, 0x0, x2, x1);
  uint32_t x20;
  fiat_p224_uint1 x21;
  fiat_p224_addcarryx_u32(&x20, &x21, x19, 0x0, 0x0);
  uint32_t x22;
  fiat_p224_uint1 x23;
  fiat_p224_addcarryx_u32(&x22, &x23, 0x0, (arg1[1]), (fiat_p224_uint1)x20);
  uint32_t x24;
  uint32_t x25;
  fiat_p224_mulx_u32(&x24, &x25, x22, UINT32_C(0xffffffff));
  uint32_t x26;
  uint32_t x27;
  fiat_p224_mulx_u32(&x26, &x27, x24, UINT32_C(0xffffffff));
  uint32_t x28;
  uint32_t x29;
  fiat_p224_mulx_u32(&x28, &x29, x24, UINT32_C(0xffffffff));
  uint32_t x30;
  uint32_t x31;
  fiat_p224_mulx_u32(&x30, &x31, x24, UINT32_C(0xffffffff));
  uint32_t x32;
  uint32_t x33;
  fiat_p224_mulx_u32(&x32, &x33, x24, UINT32_C(0xffffffff));
  uint32_t x34;
  fiat_p224_uint1 x35;
  fiat_p224_addcarryx_u32(&x34, &x35, 0x0, x30, x33);
  uint32_t x36;
  fiat_p224_uint1 x37;
  fiat_p224_addcarryx_u32(&x36, &x37, x35, x28, x31);
  uint32_t x38;
  fiat_p224_uint1 x39;
  fiat_p224_addcarryx_u32(&x38, &x39, x37, x26, x29);
  uint32_t x40;
  fiat_p224_uint1 x41;
  fiat_p224_addcarryx_u32(&x40, &x41, x23, 0x0, 0x0);
  uint32_t x42;
  fiat_p224_uint1 x43;
  fiat_p224_addcarryx_u32(&x42, &x43, 0x0, x24, x22);
  uint32_t x44;
  fiat_p224_uint1 x45;
  fiat_p224_addcarryx_u32(&x44, &x45, x43, 0x0, (fiat_p224_uint1)x40);
  uint32_t x46;
  fiat_p224_uint1 x47;
  fiat_p224_addcarryx_u32(&x46, &x47, x45, 0x0, x10);
  uint32_t x48;
  fiat_p224_uint1 x49;
  fiat_p224_addcarryx_u32(&x48, &x49, x47, x32, x12);
  uint32_t x50;
  fiat_p224_uint1 x51;
  fiat_p224_addcarryx_u32(&x50, &x51, x49, x34, x14);
  uint32_t x52;
  fiat_p224_uint1 x53;
  fiat_p224_addcarryx_u32(&x52, &x53, x51, x36, x16);
  uint32_t x54;
  fiat_p224_uint1 x55;
  fiat_p224_addcarryx_u32(&x54, &x55, x17, 0x0, x5);
  uint32_t x56;
  fiat_p224_uint1 x57;
  fiat_p224_addcarryx_u32(&x56, &x57, x53, x38, x54);
  uint32_t x58;
  fiat_p224_uint1 x59;
  fiat_p224_addcarryx_u32(&x58, &x59, x39, 0x0, x27);
  uint32_t x60;
  fiat_p224_uint1 x61;
  fiat_p224_addcarryx_u32(&x60, &x61, x57, x58, 0x0);
  uint32_t x62;
  fiat_p224_uint1 x63;
  fiat_p224_addcarryx_u32(&x62, &x63, 0x0, (arg1[2]), x44);
  uint32_t x64;
  fiat_p224_uint1 x65;
  fiat_p224_addcarryx_u32(&x64, &x65, x63, 0x0, x46);
  uint32_t x66;
  fiat_p224_uint1 x67;
  fiat_p224_addcarryx_u32(&x66, &x67, x65, 0x0, x48);
  uint32_t x68;
  fiat_p224_uint1 x69;
  fiat_p224_addcarryx_u32(&x68, &x69, x67, 0x0, x50);
  uint32_t x70;
  fiat_p224_uint1 x71;
  fiat_p224_addcarryx_u32(&x70, &x71, x69, 0x0, x52);
  uint32_t x72;
  fiat_p224_uint1 x73;
  fiat_p224_addcarryx_u32(&x72, &x73, x71, 0x0, x56);
  uint32_t x74;
  fiat_p224_uint1 x75;
  fiat_p224_addcarryx_u32(&x74, &x75, x73, 0x0, x60);
  uint32_t x76;
  fiat_p224_uint1 x77;
  fiat_p224_addcarryx_u32(&x76, &x77, x61, 0x0, 0x0);
  uint32_t x78;
  fiat_p224_uint1 x79;
  fiat_p224_addcarryx_u32(&x78, &x79, x75, 0x0, (fiat_p224_uint1)x76);
  uint32_t x80;
  uint32_t x81;
  fiat_p224_mulx_u32(&x80, &x81, x62, UINT32_C(0xffffffff));
  uint32_t x82;
  uint32_t x83;
  fiat_p224_mulx_u32(&x82, &x83, x80, UINT32_C(0xffffffff));
  uint32_t x84;
  uint32_t x85;
  fiat_p224_mulx_u32(&x84, &x85, x80, UINT32_C(0xffffffff));
  uint32_t x86;
  uint32_t x87;
  fiat_p224_mulx_u32(&x86, &x87, x80, UINT32_C(0xffffffff));
  uint32_t x88;
  uint32_t x89;
  fiat_p224_mulx_u32(&x88, &x89, x80, UINT32_C(0xffffffff));
  uint32_t x90;
  fiat_p224_uint1 x91;
  fiat_p224_addcarryx_u32(&x90, &x91, 0x0, x86, x89);
  uint32_t x92;
  fiat_p224_uint1 x93;
  fiat_p224_addcarryx_u32(&x92, &x93, x91, x84, x87);
  uint32_t x94;
  fiat_p224_uint1 x95;
  fiat_p224_addcarryx_u32(&x94, &x95, x93, x82, x85);
  uint32_t x96;
  fiat_p224_uint1 x97;
  fiat_p224_addcarryx_u32(&x96, &x97, 0x0, x80, x62);
  uint32_t x98;
  fiat_p224_uint1 x99;
  fiat_p224_addcarryx_u32(&x98, &x99, x97, 0x0, x64);
  uint32_t x100;
  fiat_p224_uint1 x101;
  fiat_p224_addcarryx_u32(&x100, &x101, x99, 0x0, x66);
  uint32_t x102;
  fiat_p224_uint1 x103;
  fiat_p224_addcarryx_u32(&x102, &x103, x101, x88, x68);
  uint32_t x104;
  fiat_p224_uint1 x105;
  fiat_p224_addcarryx_u32(&x104, &x105, x103, x90, x70);
  uint32_t x106;
  fiat_p224_uint1 x107;
  fiat_p224_addcarryx_u32(&x106, &x107, x105, x92, x72);
  uint32_t x108;
  fiat_p224_uint1 x109;
  fiat_p224_addcarryx_u32(&x108, &x109, x107, x94, x74);
  uint32_t x110;
  fiat_p224_uint1 x111;
  fiat_p224_addcarryx_u32(&x110, &x111, x95, 0x0, x83);
  uint32_t x112;
  fiat_p224_uint1 x113;
  fiat_p224_addcarryx_u32(&x112, &x113, x109, x110, x78);
  uint32_t x114;
  fiat_p224_uint1 x115;
  fiat_p224_addcarryx_u32(&x114, &x115, 0x0, (arg1[3]), x98);
  uint32_t x116;
  fiat_p224_uint1 x117;
  fiat_p224_addcarryx_u32(&x116, &x117, x115, 0x0, x100);
  uint32_t x118;
  fiat_p224_uint1 x119;
  fiat_p224_addcarryx_u32(&x118, &x119, x117, 0x0, x102);
  uint32_t x120;
  fiat_p224_uint1 x121;
  fiat_p224_addcarryx_u32(&x120, &x121, x119, 0x0, x104);
  uint32_t x122;
  fiat_p224_uint1 x123;
  fiat_p224_addcarryx_u32(&x122, &x123, x121, 0x0, x106);
  uint32_t x124;
  fiat_p224_uint1 x125;
  fiat_p224_addcarryx_u32(&x124, &x125, x123, 0x0, x108);
  uint32_t x126;
  fiat_p224_uint1 x127;
  fiat_p224_addcarryx_u32(&x126, &x127, x125, 0x0, x112);
  uint32_t x128;
  fiat_p224_uint1 x129;
  fiat_p224_addcarryx_u32(&x128, &x129, x113, 0x0, x79);
  uint32_t x130;
  fiat_p224_uint1 x131;
  fiat_p224_addcarryx_u32(&x130, &x131, x127, 0x0, (fiat_p224_uint1)x128);
  uint32_t x132;
  uint32_t x133;
  fiat_p224_mulx_u32(&x132, &x133, x114, UINT32_C(0xffffffff));
  uint32_t x134;
  uint32_t x135;
  fiat_p224_mulx_u32(&x134, &x135, x132, UINT32_C(0xffffffff));
  uint32_t x136;
  uint32_t x137;
  fiat_p224_mulx_u32(&x136, &x137, x132, UINT32_C(0xffffffff));
  uint32_t x138;
  uint32_t x139;
  fiat_p224_mulx_u32(&x138, &x139, x132, UINT32_C(0xffffffff));
  uint32_t x140;
  uint32_t x141;
  fiat_p224_mulx_u32(&x140, &x141, x132, UINT32_C(0xffffffff));
  uint32_t x142;
  fiat_p224_uint1 x143;
  fiat_p224_addcarryx_u32(&x142, &x143, 0x0, x138, x141);
  uint32_t x144;
  fiat_p224_uint1 x145;
  fiat_p224_addcarryx_u32(&x144, &x145, x143, x136, x139);
  uint32_t x146;
  fiat_p224_uint1 x147;
  fiat_p224_addcarryx_u32(&x146, &x147, x145, x134, x137);
  uint32_t x148;
  fiat_p224_uint1 x149;
  fiat_p224_addcarryx_u32(&x148, &x149, 0x0, x132, x114);
  uint32_t x150;
  fiat_p224_uint1 x151;
  fiat_p224_addcarryx_u32(&x150, &x151, x149, 0x0, x116);
  uint32_t x152;
  fiat_p224_uint1 x153;
  fiat_p224_addcarryx_u32(&x152, &x153, x151, 0x0, x118);
  uint32_t x154;
  fiat_p224_uint1 x155;
  fiat_p224_addcarryx_u32(&x154, &x155, x153, x140, x120);
  uint32_t x156;
  fiat_p224_uint1 x157;
  fiat_p224_addcarryx_u32(&x156, &x157, x155, x142, x122);
  uint32_t x158;
  fiat_p224_uint1 x159;
  fiat_p224_addcarryx_u32(&x158, &x159, x157, x144, x124);
  uint32_t x160;
  fiat_p224_uint1 x161;
  fiat_p224_addcarryx_u32(&x160, &x161, x159, x146, x126);
  uint32_t x162;
  fiat_p224_uint1 x163;
  fiat_p224_addcarryx_u32(&x162, &x163, x147, 0x0, x135);
  uint32_t x164;
  fiat_p224_uint1 x165;
  fiat_p224_addcarryx_u32(&x164, &x165, x161, x162, x130);
  uint32_t x166;
  fiat_p224_uint1 x167;
  fiat_p224_addcarryx_u32(&x166, &x167, 0x0, (arg1[4]), x150);
  uint32_t x168;
  fiat_p224_uint1 x169;
  fiat_p224_addcarryx_u32(&x168, &x169, x167, 0x0, x152);
  uint32_t x170;
  fiat_p224_uint1 x171;
  fiat_p224_addcarryx_u32(&x170, &x171, x169, 0x0, x154);
  uint32_t x172;
  fiat_p224_uint1 x173;
  fiat_p224_addcarryx_u32(&x172, &x173, x171, 0x0, x156);
  uint32_t x174;
  fiat_p224_uint1 x175;
  fiat_p224_addcarryx_u32(&x174, &x175, x173, 0x0, x158);
  uint32_t x176;
  fiat_p224_uint1 x177;
  fiat_p224_addcarryx_u32(&x176, &x177, x175, 0x0, x160);
  uint32_t x178;
  fiat_p224_uint1 x179;
  fiat_p224_addcarryx_u32(&x178, &x179, x177, 0x0, x164);
  uint32_t x180;
  fiat_p224_uint1 x181;
  fiat_p224_addcarryx_u32(&x180, &x181, x165, 0x0, x131);
  uint32_t x182;
  fiat_p224_uint1 x183;
  fiat_p224_addcarryx_u32(&x182, &x183, x179, 0x0, (fiat_p224_uint1)x180);
  uint32_t x184;
  uint32_t x185;
  fiat_p224_mulx_u32(&x184, &x185, x166, UINT32_C(0xffffffff));
  uint32_t x186;
  uint32_t x187;
  fiat_p224_mulx_u32(&x186, &x187, x184, UINT32_C(0xffffffff));
  uint32_t x188;
  uint32_t x189;
  fiat_p224_mulx_u32(&x188, &x189, x184, UINT32_C(0xffffffff));
  uint32_t x190;
  uint32_t x191;
  fiat_p224_mulx_u32(&x190, &x191, x184, UINT32_C(0xffffffff));
  uint32_t x192;
  uint32_t x193;
  fiat_p224_mulx_u32(&x192, &x193, x184, UINT32_C(0xffffffff));
  uint32_t x194;
  fiat_p224_uint1 x195;
  fiat_p224_addcarryx_u32(&x194, &x195, 0x0, x190, x193);
  uint32_t x196;
  fiat_p224_uint1 x197;
  fiat_p224_addcarryx_u32(&x196, &x197, x195, x188, x191);
  uint32_t x198;
  fiat_p224_uint1 x199;
  fiat_p224_addcarryx_u32(&x198, &x199, x197, x186, x189);
  uint32_t x200;
  fiat_p224_uint1 x201;
  fiat_p224_addcarryx_u32(&x200, &x201, 0x0, x184, x166);
  uint32_t x202;
  fiat_p224_uint1 x203;
  fiat_p224_addcarryx_u32(&x202, &x203, x201, 0x0, x168);
  uint32_t x204;
  fiat_p224_uint1 x205;
  fiat_p224_addcarryx_u32(&x204, &x205, x203, 0x0, x170);
  uint32_t x206;
  fiat_p224_uint1 x207;
  fiat_p224_addcarryx_u32(&x206, &x207, x205, x192, x172);
  uint32_t x208;
  fiat_p224_uint1 x209;
  fiat_p224_addcarryx_u32(&x208, &x209, x207, x194, x174);
  uint32_t x210;
  fiat_p224_uint1 x211;
  fiat_p224_addcarryx_u32(&x210, &x211, x209, x196, x176);
  uint32_t x212;
  fiat_p224_uint1 x213;
  fiat_p224_addcarryx_u32(&x212, &x213, x211, x198, x178);
  uint32_t x214;
  fiat_p224_uint1 x215;
  fiat_p224_addcarryx_u32(&x214, &x215, x199, 0x0, x187);
  uint32_t x216;
  fiat_p224_uint1 x217;
  fiat_p224_addcarryx_u32(&x216, &x217, x213, x214, x182);
  uint32_t x218;
  fiat_p224_uint1 x219;
  fiat_p224_addcarryx_u32(&x218, &x219, 0x0, (arg1[5]), x202);
  uint32_t x220;
  fiat_p224_uint1 x221;
  fiat_p224_addcarryx_u32(&x220, &x221, x219, 0x0, x204);
  uint32_t x222;
  fiat_p224_uint1 x223;
  fiat_p224_addcarryx_u32(&x222, &x223, x221, 0x0, x206);
  uint32_t x224;
  fiat_p224_uint1 x225;
  fiat_p224_addcarryx_u32(&x224, &x225, x223, 0x0, x208);
  uint32_t x226;
  fiat_p224_uint1 x227;
  fiat_p224_addcarryx_u32(&x226, &x227, x225, 0x0, x210);
  uint32_t x228;
  fiat_p224_uint1 x229;
  fiat_p224_addcarryx_u32(&x228, &x229, x227, 0x0, x212);
  uint32_t x230;
  fiat_p224_uint1 x231;
  fiat_p224_addcarryx_u32(&x230, &x231, x229, 0x0, x216);
  uint32_t x232;
  fiat_p224_uint1 x233;
  fiat_p224_addcarryx_u32(&x232, &x233, x217, 0x0, x183);
  uint32_t x234;
  fiat_p224_uint1 x235;
  fiat_p224_addcarryx_u32(&x234, &x235, x231, 0x0, (fiat_p224_uint1)x232);
  uint32_t x236;
  uint32_t x237;
  fiat_p224_mulx_u32(&x236, &x237, x218, UINT32_C(0xffffffff));
  uint32_t x238;
  uint32_t x239;
  fiat_p224_mulx_u32(&x238, &x239, x236, UINT32_C(0xffffffff));
  uint32_t x240;
  uint32_t x241;
  fiat_p224_mulx_u32(&x240, &x241, x236, UINT32_C(0xffffffff));
  uint32_t x242;
  uint32_t x243;
  fiat_p224_mulx_u32(&x242, &x243, x236, UINT32_C(0xffffffff));
  uint32_t x244;
  uint32_t x245;
  fiat_p224_mulx_u32(&x244, &x245, x236, UINT32_C(0xffffffff));
  uint32_t x246;
  fiat_p224_uint1 x247;
  fiat_p224_addcarryx_u32(&x246, &x247, 0x0, x242, x245);
  uint32_t x248;
  fiat_p224_uint1 x249;
  fiat_p224_addcarryx_u32(&x248, &x249, x247, x240, x243);
  uint32_t x250;
  fiat_p224_uint1 x251;
  fiat_p224_addcarryx_u32(&x250, &x251, x249, x238, x241);
  uint32_t x252;
  fiat_p224_uint1 x253;
  fiat_p224_addcarryx_u32(&x252, &x253, 0x0, x236, x218);
  uint32_t x254;
  fiat_p224_uint1 x255;
  fiat_p224_addcarryx_u32(&x254, &x255, x253, 0x0, x220);
  uint32_t x256;
  fiat_p224_uint1 x257;
  fiat_p224_addcarryx_u32(&x256, &x257, x255, 0x0, x222);
  uint32_t x258;
  fiat_p224_uint1 x259;
  fiat_p224_addcarryx_u32(&x258, &x259, x257, x244, x224);
  uint32_t x260;
  fiat_p224_uint1 x261;
  fiat_p224_addcarryx_u32(&x260, &x261, x259, x246, x226);
  uint32_t x262;
  fiat_p224_uint1 x263;
  fiat_p224_addcarryx_u32(&x262, &x263, x261, x248, x228);
  uint32_t x264;
  fiat_p224_uint1 x265;
  fiat_p224_addcarryx_u32(&x264, &x265, x263, x250, x230);
  uint32_t x266;
  fiat_p224_uint1 x267;
  fiat_p224_addcarryx_u32(&x266, &x267, x251, 0x0, x239);
  uint32_t x268;
  fiat_p224_uint1 x269;
  fiat_p224_addcarryx_u32(&x268, &x269, x265, x266, x234);
  uint32_t x270;
  fiat_p224_uint1 x271;
  fiat_p224_addcarryx_u32(&x270, &x271, 0x0, (arg1[6]), x254);
  uint32_t x272;
  fiat_p224_uint1 x273;
  fiat_p224_addcarryx_u32(&x272, &x273, x271, 0x0, x256);
  uint32_t x274;
  fiat_p224_uint1 x275;
  fiat_p224_addcarryx_u32(&x274, &x275, x273, 0x0, x258);
  uint32_t x276;
  fiat_p224_uint1 x277;
  fiat_p224_addcarryx_u32(&x276, &x277, x275, 0x0, x260);
  uint32_t x278;
  fiat_p224_uint1 x279;
  fiat_p224_addcarryx_u32(&x278, &x279, x277, 0x0, x262);
  uint32_t x280;
  fiat_p224_uint1 x281;
  fiat_p224_addcarryx_u32(&x280, &x281, x279, 0x0, x264);
  uint32_t x282;
  fiat_p224_uint1 x283;
  fiat_p224_addcarryx_u32(&x282, &x283, x281, 0x0, x268);
  uint32_t x284;
  fiat_p224_uint1 x285;
  fiat_p224_addcarryx_u32(&x284, &x285, x269, 0x0, x235);
  uint32_t x286;
  fiat_p224_uint1 x287;
  fiat_p224_addcarryx_u32(&x286, &x287, x283, 0x0, (fiat_p224_uint1)x284);
  uint32_t x288;
  uint32_t x289;
  fiat_p224_mulx_u32(&x288, &x289, x270, UINT32_C(0xffffffff));
  uint32_t x290;
  uint32_t x291;
  fiat_p224_mulx_u32(&x290, &x291, x288, UINT32_C(0xffffffff));
  uint32_t x292;
  uint32_t x293;
  fiat_p224_mulx_u32(&x292, &x293, x288, UINT32_C(0xffffffff));
  uint32_t x294;
  uint32_t x295;
  fiat_p224_mulx_u32(&x294, &x295, x288, UINT32_C(0xffffffff));
  uint32_t x296;
  uint32_t x297;
  fiat_p224_mulx_u32(&x296, &x297, x288, UINT32_C(0xffffffff));
  uint32_t x298;
  fiat_p224_uint1 x299;
  fiat_p224_addcarryx_u32(&x298, &x299, 0x0, x294, x297);
  uint32_t x300;
  fiat_p224_uint1 x301;
  fiat_p224_addcarryx_u32(&x300, &x301, x299, x292, x295);
  uint32_t x302;
  fiat_p224_uint1 x303;
  fiat_p224_addcarryx_u32(&x302, &x303, x301, x290, x293);
  uint32_t x304;
  fiat_p224_uint1 x305;
  fiat_p224_addcarryx_u32(&x304, &x305, 0x0, x288, x270);
  uint32_t x306;
  fiat_p224_uint1 x307;
  fiat_p224_addcarryx_u32(&x306, &x307, x305, 0x0, x272);
  uint32_t x308;
  fiat_p224_uint1 x309;
  fiat_p224_addcarryx_u32(&x308, &x309, x307, 0x0, x274);
  uint32_t x310;
  fiat_p224_uint1 x311;
  fiat_p224_addcarryx_u32(&x310, &x311, x309, x296, x276);
  uint32_t x312;
  fiat_p224_uint1 x313;
  fiat_p224_addcarryx_u32(&x312, &x313, x311, x298, x278);
  uint32_t x314;
  fiat_p224_uint1 x315;
  fiat_p224_addcarryx_u32(&x314, &x315, x313, x300, x280);
  uint32_t x316;
  fiat_p224_uint1 x317;
  fiat_p224_addcarryx_u32(&x316, &x317, x315, x302, x282);
  uint32_t x318;
  fiat_p224_uint1 x319;
  fiat_p224_addcarryx_u32(&x318, &x319, x303, 0x0, x291);
  uint32_t x320;
  fiat_p224_uint1 x321;
  fiat_p224_addcarryx_u32(&x320, &x321, x317, x318, x286);
  uint32_t x322;
  fiat_p224_uint1 x323;
  fiat_p224_subborrowx_u32(&x322, &x323, 0x0, x306, 0x1);
  uint32_t x324;
  fiat_p224_uint1 x325;
  fiat_p224_subborrowx_u32(&x324, &x325, x323, x308, 0x0);
  uint32_t x326;
  fiat_p224_uint1 x327;
  fiat_p224_subborrowx_u32(&x326, &x327, x325, x310, 0x0);
  uint32_t x328;
  fiat_p224_uint1 x329;
  fiat_p224_subborrowx_u32(&x328, &x329, x327, x312, UINT32_C(0xffffffff));
  uint32_t x330;
  fiat_p224_uint1 x331;
  fiat_p224_subborrowx_u32(&x330, &x331, x329, x314, UINT32_C(0xffffffff));
  uint32_t x332;
  fiat_p224_uint1 x333;
  fiat_p224_subborrowx_u32(&x332, &x333, x331, x316, UINT32_C(0xffffffff));
  uint32_t x334;
  fiat_p224_uint1 x335;
  fiat_p224_subborrowx_u32(&x334, &x335, x333, x320, UINT32_C(0xffffffff));
  uint32_t x336;
  fiat_p224_uint1 x337;
  fiat_p224_addcarryx_u32(&x336, &x337, x321, 0x0, x287);
  uint32_t x338;
  fiat_p224_uint1 x339;
  fiat_p224_subborrowx_u32(&x338, &x339, x335, (fiat_p224_uint1)x336, 0x0);
  uint32_t x340;
  fiat_p224_cmovznz_u32(&x340, x339, x322, x306);
  uint32_t x341;
  fiat_p224_cmovznz_u32(&x341, x339, x324, x308);
  uint32_t x342;
  fiat_p224_cmovznz_u32(&x342, x339, x326, x310);
  uint32_t x343;
  fiat_p224_cmovznz_u32(&x343, x339, x328, x312);
  uint32_t x344;
  fiat_p224_cmovznz_u32(&x344, x339, x330, x314);
  uint32_t x345;
  fiat_p224_cmovznz_u32(&x345, x339, x332, x316);
  uint32_t x346;
  fiat_p224_cmovznz_u32(&x346, x339, x334, x320);
  out1[0] = x340;
  out1[1] = x341;
  out1[2] = x342;
  out1[3] = x343;
  out1[4] = x344;
  out1[5] = x345;
  out1[6] = x346;
}

/*
 * Input Bounds:
 *   arg1: [[0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff]]
 * Output Bounds:
 *   out1: [0x0 ~> 0xffffffff]
 */
static void fiat_p224_nonzero(uint32_t* out1, const uint32_t arg1[7]) {
  uint32_t x1 = ((arg1[0]) | ((arg1[1]) | ((arg1[2]) | ((arg1[3]) | ((arg1[4]) | ((arg1[5]) | ((arg1[6]) | (uint32_t)0x0)))))));
  *out1 = x1;
}

/*
 * Input Bounds:
 *   arg1: [0x0 ~> 0x1]
 *   arg2: [[0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff]]
 *   arg3: [[0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff]]
 * Output Bounds:
 *   out1: [[0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff]]
 */
static void fiat_p224_selectznz(uint32_t out1[7], fiat_p224_uint1 arg1, const uint32_t arg2[7], const uint32_t arg3[7]) {
  uint32_t x1;
  fiat_p224_cmovznz_u32(&x1, arg1, (arg2[0]), (arg3[0]));
  uint32_t x2;
  fiat_p224_cmovznz_u32(&x2, arg1, (arg2[1]), (arg3[1]));
  uint32_t x3;
  fiat_p224_cmovznz_u32(&x3, arg1, (arg2[2]), (arg3[2]));
  uint32_t x4;
  fiat_p224_cmovznz_u32(&x4, arg1, (arg2[3]), (arg3[3]));
  uint32_t x5;
  fiat_p224_cmovznz_u32(&x5, arg1, (arg2[4]), (arg3[4]));
  uint32_t x6;
  fiat_p224_cmovznz_u32(&x6, arg1, (arg2[5]), (arg3[5]));
  uint32_t x7;
  fiat_p224_cmovznz_u32(&x7, arg1, (arg2[6]), (arg3[6]));
  out1[0] = x1;
  out1[1] = x2;
  out1[2] = x3;
  out1[3] = x4;
  out1[4] = x5;
  out1[5] = x6;
  out1[6] = x7;
}

/*
 * Input Bounds:
 *   arg1: [[0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff]]
 * Output Bounds:
 *   out1: [[0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff]]
 */
static void fiat_p224_to_bytes(uint8_t out1[28], const uint32_t arg1[7]) {
  uint32_t x1 = (arg1[6]);
  uint32_t x2 = (arg1[5]);
  uint32_t x3 = (arg1[4]);
  uint32_t x4 = (arg1[3]);
  uint32_t x5 = (arg1[2]);
  uint32_t x6 = (arg1[1]);
  uint32_t x7 = (arg1[0]);
  uint32_t x8 = (x7 >> 8);
  uint8_t x9 = (uint8_t)(x7 & UINT8_C(0xff));
  uint32_t x10 = (x8 >> 8);
  uint8_t x11 = (uint8_t)(x8 & UINT8_C(0xff));
  uint8_t x12 = (uint8_t)(x10 >> 8);
  uint8_t x13 = (uint8_t)(x10 & UINT8_C(0xff));
  fiat_p224_uint1 x14 = (fiat_p224_uint1)(x12 >> 8);
  uint8_t x15 = (uint8_t)(x12 & UINT8_C(0xff));
  uint32_t x16 = (x14 + x6);
  uint32_t x17 = (x16 >> 8);
  uint8_t x18 = (uint8_t)(x16 & UINT8_C(0xff));
  uint32_t x19 = (x17 >> 8);
  uint8_t x20 = (uint8_t)(x17 & UINT8_C(0xff));
  uint8_t x21 = (uint8_t)(x19 >> 8);
  uint8_t x22 = (uint8_t)(x19 & UINT8_C(0xff));
  fiat_p224_uint1 x23 = (fiat_p224_uint1)(x21 >> 8);
  uint8_t x24 = (uint8_t)(x21 & UINT8_C(0xff));
  uint32_t x25 = (x23 + x5);
  uint32_t x26 = (x25 >> 8);
  uint8_t x27 = (uint8_t)(x25 & UINT8_C(0xff));
  uint32_t x28 = (x26 >> 8);
  uint8_t x29 = (uint8_t)(x26 & UINT8_C(0xff));
  uint8_t x30 = (uint8_t)(x28 >> 8);
  uint8_t x31 = (uint8_t)(x28 & UINT8_C(0xff));
  fiat_p224_uint1 x32 = (fiat_p224_uint1)(x30 >> 8);
  uint8_t x33 = (uint8_t)(x30 & UINT8_C(0xff));
  uint32_t x34 = (x32 + x4);
  uint32_t x35 = (x34 >> 8);
  uint8_t x36 = (uint8_t)(x34 & UINT8_C(0xff));
  uint32_t x37 = (x35 >> 8);
  uint8_t x38 = (uint8_t)(x35 & UINT8_C(0xff));
  uint8_t x39 = (uint8_t)(x37 >> 8);
  uint8_t x40 = (uint8_t)(x37 & UINT8_C(0xff));
  fiat_p224_uint1 x41 = (fiat_p224_uint1)(x39 >> 8);
  uint8_t x42 = (uint8_t)(x39 & UINT8_C(0xff));
  uint32_t x43 = (x41 + x3);
  uint32_t x44 = (x43 >> 8);
  uint8_t x45 = (uint8_t)(x43 & UINT8_C(0xff));
  uint32_t x46 = (x44 >> 8);
  uint8_t x47 = (uint8_t)(x44 & UINT8_C(0xff));
  uint8_t x48 = (uint8_t)(x46 >> 8);
  uint8_t x49 = (uint8_t)(x46 & UINT8_C(0xff));
  fiat_p224_uint1 x50 = (fiat_p224_uint1)(x48 >> 8);
  uint8_t x51 = (uint8_t)(x48 & UINT8_C(0xff));
  uint32_t x52 = (x50 + x2);
  uint32_t x53 = (x52 >> 8);
  uint8_t x54 = (uint8_t)(x52 & UINT8_C(0xff));
  uint32_t x55 = (x53 >> 8);
  uint8_t x56 = (uint8_t)(x53 & UINT8_C(0xff));
  uint8_t x57 = (uint8_t)(x55 >> 8);
  uint8_t x58 = (uint8_t)(x55 & UINT8_C(0xff));
  fiat_p224_uint1 x59 = (fiat_p224_uint1)(x57 >> 8);
  uint8_t x60 = (uint8_t)(x57 & UINT8_C(0xff));
  uint32_t x61 = (x59 + x1);
  uint32_t x62 = (x61 >> 8);
  uint8_t x63 = (uint8_t)(x61 & UINT8_C(0xff));
  uint32_t x64 = (x62 >> 8);
  uint8_t x65 = (uint8_t)(x62 & UINT8_C(0xff));
  uint8_t x66 = (uint8_t)(x64 >> 8);
  uint8_t x67 = (uint8_t)(x64 & UINT8_C(0xff));
  out1[0] = x9;
  out1[1] = x11;
  out1[2] = x13;
  out1[3] = x15;
  out1[4] = x18;
  out1[5] = x20;
  out1[6] = x22;
  out1[7] = x24;
  out1[8] = x27;
  out1[9] = x29;
  out1[10] = x31;
  out1[11] = x33;
  out1[12] = x36;
  out1[13] = x38;
  out1[14] = x40;
  out1[15] = x42;
  out1[16] = x45;
  out1[17] = x47;
  out1[18] = x49;
  out1[19] = x51;
  out1[20] = x54;
  out1[21] = x56;
  out1[22] = x58;
  out1[23] = x60;
  out1[24] = x63;
  out1[25] = x65;
  out1[26] = x67;
  out1[27] = x66;
}

/*
 * Input Bounds:
 *   arg1: [[0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff], [0x0 ~> 0xff]]
 * Output Bounds:
 *   out1: [[0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff], [0x0 ~> 0xffffffff]]
 */
static void fiat_p224_from_bytes(uint32_t out1[7], const uint8_t arg1[28]) {
  uint32_t x1 = ((uint32_t)(arg1[27]) << 24);
  uint32_t x2 = ((uint32_t)(arg1[26]) << 16);
  uint32_t x3 = ((uint32_t)(arg1[25]) << 8);
  uint8_t x4 = (arg1[24]);
  uint32_t x5 = ((uint32_t)(arg1[23]) << 24);
  uint32_t x6 = ((uint32_t)(arg1[22]) << 16);
  uint32_t x7 = ((uint32_t)(arg1[21]) << 8);
  uint8_t x8 = (arg1[20]);
  uint32_t x9 = ((uint32_t)(arg1[19]) << 24);
  uint32_t x10 = ((uint32_t)(arg1[18]) << 16);
  uint32_t x11 = ((uint32_t)(arg1[17]) << 8);
  uint8_t x12 = (arg1[16]);
  uint32_t x13 = ((uint32_t)(arg1[15]) << 24);
  uint32_t x14 = ((uint32_t)(arg1[14]) << 16);
  uint32_t x15 = ((uint32_t)(arg1[13]) << 8);
  uint8_t x16 = (arg1[12]);
  uint32_t x17 = ((uint32_t)(arg1[11]) << 24);
  uint32_t x18 = ((uint32_t)(arg1[10]) << 16);
  uint32_t x19 = ((uint32_t)(arg1[9]) << 8);
  uint8_t x20 = (arg1[8]);
  uint32_t x21 = ((uint32_t)(arg1[7]) << 24);
  uint32_t x22 = ((uint32_t)(arg1[6]) << 16);
  uint32_t x23 = ((uint32_t)(arg1[5]) << 8);
  uint8_t x24 = (arg1[4]);
  uint32_t x25 = ((uint32_t)(arg1[3]) << 24);
  uint32_t x26 = ((uint32_t)(arg1[2]) << 16);
  uint32_t x27 = ((uint32_t)(arg1[1]) << 8);
  uint8_t x28 = (arg1[0]);
  uint32_t x29 = (x28 + (x27 + (x26 + x25)));
  fiat_p224_uint1 x30 = (fiat_p224_uint1)((uint64_t)x29 >> 32);
  uint32_t x31 = (x29 & UINT32_C(0xffffffff));
  uint32_t x32 = (x4 + (x3 + (x2 + x1)));
  uint32_t x33 = (x8 + (x7 + (x6 + x5)));
  uint32_t x34 = (x12 + (x11 + (x10 + x9)));
  uint32_t x35 = (x16 + (x15 + (x14 + x13)));
  uint32_t x36 = (x20 + (x19 + (x18 + x17)));
  uint32_t x37 = (x24 + (x23 + (x22 + x21)));
  uint32_t x38 = (x30 + x37);
  fiat_p224_uint1 x39 = (fiat_p224_uint1)((uint64_t)x38 >> 32);
  uint32_t x40 = (x38 & UINT32_C(0xffffffff));
  uint32_t x41 = (x39 + x36);
  fiat_p224_uint1 x42 = (fiat_p224_uint1)((uint64_t)x41 >> 32);
  uint32_t x43 = (x41 & UINT32_C(0xffffffff));
  uint32_t x44 = (x42 + x35);
  fiat_p224_uint1 x45 = (fiat_p224_uint1)((uint64_t)x44 >> 32);
  uint32_t x46 = (x44 & UINT32_C(0xffffffff));
  uint32_t x47 = (x45 + x34);
  fiat_p224_uint1 x48 = (fiat_p224_uint1)((uint64_t)x47 >> 32);
  uint32_t x49 = (x47 & UINT32_C(0xffffffff));
  uint32_t x50 = (x48 + x33);
  fiat_p224_uint1 x51 = (fiat_p224_uint1)((uint64_t)x50 >> 32);
  uint32_t x52 = (x50 & UINT32_C(0xffffffff));
  uint32_t x53 = (x51 + x32);
  out1[0] = x31;
  out1[1] = x40;
  out1[2] = x43;
  out1[3] = x46;
  out1[4] = x49;
  out1[5] = x52;
  out1[6] = x53;
}

